program offlinediag

  use netcdf
  use input_mod, only:     read_data,                open_input_files,        &
                      get_dimensions,                       axes_init,        &
              read_surf_geopotential

  use get_grid_fields_mod, only:                                              &
                    grid_fields_init,                 get_grid_fields,        &
                       surf_gradient,             zonal_spectrum_grid

  use vars

  use isentropic_mod, only:                                                   &
                     isentropic_init,             sigma_to_isentropes,        &
                isentropic_variables,                      isentrdens

  use local_utilities_mod, only:                                              &
                           sigma_avg,        interpolate_half_to_full,        &
                           mrdnl_avg,            update_sfc_frequency,        &
               residual_circulations,              buoyancy_frequency,        &
                      divide_by_psfc,            update_stat_unstable,        &
                      pot_temp_deriv,                  vrtcl_gradient,        &
                      mrdnl_gradient,                       vrtcl_int,        &
                             pdf_avg

  use constants_and_switches_mod, only:                                       &
            reference_sea_level_pres,                              cp,        &
                               rdgas,                           kappa,        &
                                  pi,                           omega,        &
                                grav,                          radius,        &
                               rvgas

  use output_mod, only:                                                       &
                  create_output_file,                 write_variables,        &
                   close_output_file


  implicit none

! ------------------------------ local variables ------------------------------

  real :: time                 ! current time
  integer :: time_counter=1    ! used to calculate frequencies based on daily data
  
  ! local dimensions of gridpoint arrays
  integer ::                                                                  &
       num_lat,            &   ! total number of latitudes
       num_lon,            &   ! total number of longitudes
       num_lev,            &   ! total number of levels
       num_times,          &   ! total number of times in file
       curr_num_times          ! number of times in current segment

  integer i,j,k

! ------------------------------ executable code -----------------------------

  call read_namelist

  call open_input_files(      DataIn,                     Data_Source,        &
                            moisture,                           ozone,        &
                          moist_tend,                          precip,        &
                             virtual,                                         &
                            UVarName,                        VVarName,        &
                          VORVarName,                      DIVVarName,        &
                         TEMPVarName,                       PSVarName,        &
                         SHumVarName,                     CondVarName,        &
                         ConvVarName,               PrecipCondVarName,        &
                   PrecipConvVarName,                  UInputFileName,        &
                      VInputFileName,                VORInputFileName,        &
                    DIVInputFileName,               TEMPInputFileName,        &
                     PSInputFileName,               SHumInputFileName,        &
                   CondInputFileName,               ConvInputFileName,        &
             PrecipCondInputFileName,         PrecipConvInputFileName)



  call get_dimensions(       num_lat,                         num_lon,        &
                             num_lev,                       num_times )

  ! initialize fields based on dimensions
  call allocate_and_initialize(                                               &
                             num_lat,                         num_lon,        &
                             num_lev )

! get axes
  call axes_init(            deg_lat,                         sin_lat,        &
                          cos_lat_xy,                      cos_lat_yz )

! read_surf_geopotential must be called before grid_fields_init
  if(.not.smooth_surface)  call read_surf_geopotential(                       &
                           SGVarName,                 SGInputFileName,        &
                   surf_geopotential )

! initialize get_grid_fields module
  call grid_fields_init(      DataIn,                     is_gaussian,        &
                            moisture,                           ozone,        &
                    moist_isentropes,                     data_source,        &
                         num_fourier,                         num_lat,        &
                             num_lon,                         num_lev,        &
                          cos_lat_xy,               surf_geopotential,        &
                               sigma,                              pk,        &
                                  bk )


! initialize isentropic module
  call isentropic_init(   PotTempMin,                      PotTempMax,        &
                         theta_coord,                    dtheta_coord )

  
  curr_num_times = num_times / num_segments
  write(*,*) "Starting time ", TimeIn
  write(*,*) 'Analyzing ', curr_num_times, ' instants.'
  
  time = TimeIn
  do while (time < TimeIn + curr_num_times)

     time = time + 1.0

     write(6,*) time

     call read_data(            time,                          u_grid,        &
                              v_grid,                        vor_grid,        &
                            div_grid,                       temp_grid,        &
                             ps_grid,                       shum_grid,        &
                          ozone_grid,                 dt_qg_cond_grid,        &
                     dt_qg_conv_grid,                precip_cond_grid,        &
                    precip_conv_grid)

     ! calculate virtual temperature consistent with the approximation 
     ! used in the fms code; for use in z, mont_x_grid, N^2, svol
     if (virtual .and. moisture) then
      virtual_temp_grid = temp_grid*(1.0+shum_grid*(rvgas/rdgas-1.0))
     else
      virtual_temp_grid = temp_grid
     endif

     call get_grid_fields(    u_grid,                          v_grid,        &
                            vor_grid,                        div_grid,        &
                           temp_grid,                       shum_grid,        &
                          ozone_grid,                                         &
                             ps_grid,                       vcos_grid,        &
                              w_grid,                        pot_temp,        &
                             theta_e,                     theta_e_sat,        &
                              p_half,                          p_full,        &
                           mont_grid,                     mont_x_grid,        &
                      hor_sfctn_grid,                      sfctn_grid,        &
                       bt_sfctn_grid,                           sigma,        &
                       bc_sfctn_spec,                   bt_sfctn_spec,        &
                       pot_temp_spec,                       conv_spec,        &
                           shum_spec,                   sat_shum_spec,        &
                           rhum_spec,                    theta_e_spec,        &
                          ozone_spec,               virtual_temp_grid,        &
                         geopot_grid,                     sat_mr_grid,        &
                       sat_shum_grid,                       rhum_grid)

     ! compute w_up                                                        
     w_up_grid      = 0.0
     w_up_freq_grid = 0.0
     where( w_grid < 0.0 )
        w_up_grid = w_grid
        w_up_freq_grid = 1.0
     endwhere
        
     call update_stat_unstable(sigma,                        pot_temp,        &
                  freq_stat_unstable )


     ! average (around latitude circle) sigma-coordinate fields (inc. surface)
     call sigma_averages

     call isentropic_variables(                                               &
                             sin_lat,                        pot_temp,        &
                              p_full,           p_half(:,:,num_lev+1),        &
                              u_grid,                          v_grid,        &
                            vor_grid,                       mont_grid,        &
                               k_sfc,                          k_asfc,        &
                            p_isentr,                                         &
                            u_isentr,                        v_isentr,        &
                      abs_vor_isentr,                                         &
                         mont_isentr )


     call update_sfc_frequency(k_sfc,                   sfc_frequency )

     ! find isentropic density implied by linear interpolation 
     ! routine sigma_to_isentropes
     call isentrdens(p_isentr, dens_isentr)

     ! compute f * geostrophic mass flux (rho_mont_x)  and
     ! f * geostrophic velocity (mont_x)
     call sigma_to_isentropes(                                                &
                            pot_temp,                          p_half,        &
                            p_isentr,                           k_sfc,        &
                         mont_x_grid,                rho_mont_x_isentr )


     where (dens_isentr .gt. 1e-6)
        mont_x_isentr = rho_mont_x_isentr * 1.0 / dens_isentr
     elsewhere
        mont_x_isentr = 0.0
     endwhere

     ! note sfctn_grid is an integral quantity and sfctn_isentr thus has
     ! no density factor (optional argument integral is .true.)
     call sigma_to_isentropes(                                                &
                            pot_temp,                          p_half,        &
                            p_isentr,                          k_asfc,        &
                          sfctn_grid,                    sfctn_isentr,        &
                               .true. )
     if(moisture) then
        call sigma_to_isentropes(                                             &
                            pot_temp,                          p_half,        &
                            p_isentr,                           k_sfc,        &
                           shum_grid,                     shum_isentr)

        call sigma_to_isentropes(                                             &
                            pot_temp,                          p_half,        &
                            p_isentr,                           k_sfc,        &
                           rhum_grid,                     rhum_isentr)


        call sigma_to_isentropes(                                             &
                            pot_temp,                          p_half,        &
                            p_isentr,                           k_sfc,        &
                       sat_shum_grid,                 sat_shum_isentr)
     
        call sigma_to_isentropes(                                             &
                            pot_temp,                          p_half,        &
                            p_isentr,                           k_sfc,        &
                    v_grid*shum_grid,                    vshum_isentr)

        call sigma_to_isentropes(                                             &
                            pot_temp,                          p_half,        &
                            p_isentr,                           k_sfc,        &
                v_grid*sat_shum_grid,                vsat_shum_isentr)
     endif

     if(ozone) then
        call sigma_to_isentropes(                                             &
                            pot_temp,                          p_half,        &
                            p_isentr,                           k_sfc,        &
                          ozone_grid,                    ozone_isentr)
                                                                                                                                     
        call sigma_to_isentropes(                                             &
                            pot_temp,                          p_half,        &
                            p_isentr,                           k_sfc,        &
                   v_grid*ozone_grid,                   vozone_isentr)
     endif


     ! calculate isentropic potential vorticity
     where (dens_isentr .gt. 1e-6)
        ipv  = abs_vor_isentr * 1.0 / dens_isentr
     elsewhere
        ipv = 0.0
     endwhere

     ! average (around latitude circle) theta-coordinate fields (inc. surface)
     call theta_averages

  enddo ! end loop over times

  call average_fields(time-TimeIn)

  call eddy_fields

  ! compute mean potential temperature gradients
  call pot_temp_deriv(                                                        &
                             deg_lat,                           sigma,        &
                        pot_temp_avg,                     ps_grid_avg,        &
                   d_dy_pot_temp_avg,             d_dsig_pot_temp_avg,        &
                   d_dp_pot_temp_avg )

  ! compute mean equivalent potential temperature gradients
  if (moisture)                                                               &
    call pot_temp_deriv(                                                      &
                               deg_lat,                           sigma,      &
                           theta_e_avg,                     ps_grid_avg,      &
                      d_dy_theta_e_avg,              d_dsig_theta_e_avg,      &
                      d_dp_theta_e_avg )

  ! get Brunt-Vaeisaelae frequency
  buoyancy_avg = buoyancy_frequency(                                          &
                          p_full_avg,                virtual_temp_avg,        &
                        pot_temp_avg,               d_dp_pot_temp_avg )


  ! generate TEM and Modified TEM circulations
  call residual_circulations(                                                 &
                           sfctn_avg,             mrdnl_eddy_heatflux,        &
                 vrtcl_eddy_heatflux,               d_dy_pot_temp_avg,        &
                 d_dsig_pot_temp_avg,                     ps_grid_avg,        &
                        TEM_res_circ,                   MTEM_res_circ )

  call create_output_file(   num_lat,                         num_lev )

  call write_variables(      num_lat,                     num_fourier,        &
                             num_lev,                     time-TimeIn )

  call close_output_file()
  
contains
  
! ----------------------------------------------------------------------------

  subroutine sigma_averages

!                            --- local variables ---

    real, dimension(size(u_grid,1),size(u_grid,2)) :: u_baro, v_baro
    real, dimension(size(u_grid,1),size(u_grid,2),size(u_grid,3)) ::         &
         ucos_grid             ! u * cosine(latitude)

!                            --- executable code ---

    p_half_zon       = mrdnl_avg(p_half)
    p_full_avg       = p_full_avg        + mrdnl_avg(p_full)
    ps_grid_avg      = ps_grid_avg       + p_half_zon(:, num_lev+1)
    z_avg            = z_avg             + sigma_avg(geopot_grid / grav, ps_grid)
    temp_avg         = temp_avg          + sigma_avg(temp_grid, ps_grid)
    virtual_temp_avg = virtual_temp_avg  + sigma_avg(virtual_temp_grid, ps_grid)
    temp_var_avg     = temp_var_avg      + sigma_avg(temp_grid**2, ps_grid)
    pot_temp_avg     = pot_temp_avg      + sigma_avg(pot_temp, ps_grid)
    pot_temp_var_avg = pot_temp_var_avg  + sigma_avg(pot_temp**2, ps_grid)
    u_avg            = u_avg             + sigma_avg(u_grid, ps_grid)
    u_var_avg        = u_var_avg         + sigma_avg(u_grid**2, ps_grid)
    v_avg            = v_avg             + sigma_avg(v_grid, ps_grid)
    vort_avg         = vort_avg          + sigma_avg(vor_grid, ps_grid)
    v_var_avg        = v_var_avg         + sigma_avg(v_grid**2, ps_grid)
    w_var_avg        = w_var_avg         + sigma_avg(w_grid**2, ps_grid)
    w_cube_avg       = w_cube_avg        + sigma_avg(w_grid**3, ps_grid)

    if(moisture) then
       shum_avg         = shum_avg           + sigma_avg(shum_grid, ps_grid)
       shum_var_avg     = shum_var_avg       + sigma_avg(shum_grid**2, ps_grid)
       sat_shum_var_avg = sat_shum_var_avg   + sigma_avg(sat_shum_grid**2, ps_grid)
       ! note no cosine factor in the following
       vshum_avg        = vshum_avg          + sigma_avg(shum_grid*v_grid, ps_grid)
       wshum_avg        = wshum_avg          + sigma_avg(shum_grid*w_grid, ps_grid)
       temp_shum_avg    = temp_shum_avg      + sigma_avg(shum_grid*temp_grid, ps_grid)
       rhum_avg         = rhum_avg           + sigma_avg(rhum_grid, ps_grid)
       sat_shum_avg     = sat_shum_avg       + sigma_avg(sat_shum_grid, ps_grid) 
       theta_e_avg      = theta_e_avg        + sigma_avg(theta_e, ps_grid)
       theta_e_sat_avg  = theta_e_sat_avg    + sigma_avg(theta_e_sat, ps_grid)
       theta_e_var_avg  = theta_e_var_avg    + sigma_avg(theta_e**2, ps_grid)

       rhum_pdf        = rhum_pdf         + pdf_avg(rhum_grid, ps_grid, rhum_bin)
    endif

    if (moist_tend) then
       dt_qg_cond_avg   = dt_qg_cond_avg + sigma_avg(dt_qg_cond_grid, ps_grid)
       dt_qg_conv_avg   = dt_qg_conv_avg + sigma_avg(dt_qg_conv_grid, ps_grid)

       dt_qg_cond_counter = 0
       where(dt_qg_cond_grid.ne.0)
        dt_qg_cond_counter = 1
       endwhere

       dt_qg_conv_counter = 0
       where(dt_qg_conv_grid.ne.0)
        dt_qg_conv_counter = 1
       endwhere

       dt_qg_cond_freq   = dt_qg_cond_freq + sigma_avg(dt_qg_cond_counter, ps_grid)
       dt_qg_conv_freq   = dt_qg_conv_freq + sigma_avg(dt_qg_conv_counter, ps_grid)

    endif

    if (precip) then

       precip_cond_avg  = precip_cond_avg + sum(precip_cond_grid, 1)/num_lon
       precip_conv_avg  = precip_conv_avg + sum(precip_conv_grid, 1)/num_lon

       ! calculate precipitation intensity based on daily data
       ! (input precipitation data should be 4xday and time averaged rather than instantaneous)

       precip_cond_daily = precip_cond_daily + 0.25*precip_cond_grid
       precip_conv_daily = precip_conv_daily + 0.25*precip_conv_grid
       
       if (time_counter == 4) then  ! reached end of day
       
       
        precip_mask = precip_cond_daily .gt. precip_threshold
        precip_cond_intensity = precip_cond_intensity +                  &
          sum(precip_cond_daily, 1, mask=precip_mask)
        wet_days_cond = wet_days_cond +                                  &
          sum(one_array,1, mask=precip_mask)


        precip_mask = precip_conv_daily .gt. precip_threshold
        precip_conv_intensity = precip_conv_intensity +                  &
         sum(precip_conv_daily, 1, mask=precip_mask)
        wet_days_conv = wet_days_conv +                                  &
          sum(one_array,1, mask=precip_mask)

        precip_mask = (precip_cond_daily+precip_conv_daily) .gt. precip_threshold
        precip_intensity = precip_intensity +                            &
         sum(precip_cond_daily+precip_conv_daily, 1, mask=precip_mask)
        wet_days = wet_days +                                            &
          sum(one_array,1, mask=precip_mask)

       
        precip_cond_daily = 0
        precip_conv_daily = 0

        time_counter        = 1
       
       else
       
        time_counter = time_counter + 1
       
       endif

    endif

    if(ozone) then
       ozone_avg     = ozone_avg          + sigma_avg(ozone_grid, ps_grid)
       vozone_avg    = vozone_avg         + sigma_avg(ozone_grid*v_grid, ps_grid)
    endif


    bt_sfctn_avg     = bt_sfctn_avg      + sum(bt_sfctn_grid,1) / num_lon
    bt_sfctn_var_avg = bt_sfctn_var_avg  + sum(bt_sfctn_grid**2,1) / num_lon

    vcos_avg         = vcos_avg          + sigma_avg(vcos_grid, ps_grid)
    w_avg            = w_avg             + sigma_avg(w_grid, ps_grid)
    w_up_avg         = w_up_avg          + sigma_avg(w_up_grid, ps_grid)
    w_up_freq        = w_up_freq         + mrdnl_avg(w_up_freq_grid)
    uvcos_avg        = uvcos_avg         + sigma_avg(u_grid*vcos_grid, ps_grid)
    wu_avg           = wu_avg            + sigma_avg(u_grid*w_grid, ps_grid)
    svol_avg         = svol_avg          + sigma_avg(rdgas*virtual_temp_grid/p_full,  &
         ps_grid)
    sfctn_avg        = sfctn_avg         +                                    &
        interpolate_half_to_full(mrdnl_avg(sfctn_grid))

    hor_sfctn_avg    = hor_sfctn_avg     + mrdnl_avg(hor_sfctn_grid)
    hor_sfctn_var_avg= hor_sfctn_var_avg + mrdnl_avg(hor_sfctn_grid**2)

    surf_montgomery  = surf_geopotential + cp * temp_grid(:, :, num_lev)
    call surf_gradient(surf_montgomery, dx_surf_montgomery,                   &
         dy_surf_montgomery)
    

    ! Bernoulli function at surface
    surf_bernoulli   = surf_montgomery   +                                    &
         0.5 * (u_grid(:, :, num_lev)**2 + v_grid(:, :, num_lev)**2)

    ! compute along-surface gradient of potential temperature
    call surf_gradient(pot_temp(:,:,num_lev), dx_surf_pottemp,                &
         dy_surf_pottemp)
    
     ! y-component of surface PV flux (up to density factor)
     spv_flux_y       = - surf_bernoulli * dx_surf_pottemp     
     spv_flux_y_avg   = spv_flux_y_avg    + sum(spv_flux_y, 1) /              &
          float(num_lon)

    ! geostrophic surface flux of potential temperature (times f)
    geos_heat_flux_sfc   = geos_heat_flux_sfc +                               &
         sum(pot_temp(:,:,num_lev) * dx_surf_montgomery, 1) / float(num_lon)

    ! sigma coordinate version of mean heatfluxes
    ! meridional flux = zonal mean [ psfc * v * cos(lat) * [pot_]temp ]
    ! vertical flux   = zonal mean [ psfc * w * [pot_]temp ] with w=\dot\sigma
    mrdnl_heatflux_avg   = mrdnl_heatflux_avg +                               &
         sigma_avg(vcos_grid * pot_temp, ps_grid)
    vrtcl_heatflux_avg   = vrtcl_heatflux_avg +                               &
         sigma_avg(w_grid * pot_temp, ps_grid)
    mrdnl_tempflux_avg   = mrdnl_tempflux_avg +                               &
         sigma_avg(vcos_grid * temp_grid, ps_grid)
    vrtcl_tempflux_avg   = vrtcl_tempflux_avg +                               &
         sigma_avg(w_grid * temp_grid, ps_grid)

    mrdnl_z_flux_avg     = mrdnl_z_flux_avg +                                 &
         sigma_avg(vcos_grid * geopot_grid / grav, ps_grid)

    vrtcl_z_flux_avg     = vrtcl_z_flux_avg +                                 &
         sigma_avg(w_grid * geopot_grid / grav, ps_grid)

    if (moisture) then
      mrdnl_theta_e_flux_avg   = mrdnl_theta_e_flux_avg +                     &
           sigma_avg(vcos_grid * theta_e, ps_grid)
      vrtcl_theta_e_flux_avg   = vrtcl_theta_e_flux_avg +                     &
           sigma_avg(w_grid * theta_e, ps_grid)
    end if
    
    u_baro            = vrtcl_int(p_half, u_grid)
    v_baro            = vrtcl_int(p_half, v_grid)
    

    ! u_baro and v_baro have a surface pressure in them:
    u_baro_avg        = u_baro_avg     + mrdnl_avg(u_baro)
    v_baro_avg        = v_baro_avg     + mrdnl_avg(v_baro)
    u_baro_var_avg    = u_baro_var_avg + mrdnl_avg(u_baro**2 / ps_grid )
    v_baro_var_avg    = v_baro_var_avg + mrdnl_avg(v_baro**2 / ps_grid )

    theta_s_mont_x    = theta_s_mont_x +                                      &
         sum(pot_temp(:,:,num_lev)**2 * dx_surf_montgomery, 1) / float(num_lon)

    ! get zonal spectra
    do k = 1, num_lev
       ucos_grid(:,:,k) = u_grid(:,:,k) * cos_lat_xy
    end do

    ucos_zon_spec_avg    = ucos_zon_spec_avg  +                               &
         zonal_spectrum_grid(p_half_zon, ucos_grid  )
    vcos_zon_spec_avg    = vcos_zon_spec_avg  +                               &
         zonal_spectrum_grid(p_half_zon, vcos_grid)     
    if(moisture)                                                              &
         shum_zon_spec_avg    = shum_zon_spec_avg  +                          &
         zonal_spectrum_grid(p_half_zon, shum_grid)     
    if(ozone)                                                                 &
         ozone_zon_spec_avg    = ozone_zon_spec_avg  +                        &
         zonal_spectrum_grid(p_half_zon, ozone_grid)


 end subroutine sigma_averages
  
! ----------------------------------------------------------------------------

  subroutine theta_averages


     ! note no density weighting for p_isentr and exner
     p_isavg         = p_isavg      + mrdnl_avg(p_isentr)
     p_var_isavg     = p_var_isavg  + mrdnl_avg(p_isentr**2)
     exner_isavg     = exner_isavg  +                                         &
          cp *  mrdnl_avg((p_isentr/reference_sea_level_pres)**kappa ) 
     pexner_isavg    = pexner_isavg +                                         &
          mrdnl_avg((p_isentr/reference_sea_level_pres)**(kappa+1.0) ) 

     dens_isavg        = dens_isavg        + mrdnl_avg(dens_isentr)
     v_isavg           = v_isavg           + mrdnl_avg(dens_isentr * v_isentr)
     u_isavg           = u_isavg           + mrdnl_avg(dens_isentr * u_isentr)
     v2_isavg          = v2_isavg          + mrdnl_avg(dens_isentr * v_isentr**2)
     u2_isavg          = u2_isavg          + mrdnl_avg(dens_isentr * u_isentr**2)

     ! sfctn_isentr has no density weighting (see above) and nor
     ! should sfctn_isavg
     sfctn_isavg       = sfctn_isavg       + mrdnl_avg(sfctn_isentr)

     if(moisture)   then
     ! note linearly interpolated quantities already include density weighting
        shum_isavg      = shum_isavg   + mrdnl_avg(shum_isentr)
        rhum_isavg      = rhum_isavg   + mrdnl_avg(rhum_isentr)
        sat_shum_isavg  = sat_shum_isavg + mrdnl_avg(sat_shum_isentr)
        vshum_isavg     = vshum_isavg  + mrdnl_avg(vshum_isentr)
        vsat_shum_isavg = vsat_shum_isavg  + mrdnl_avg(vsat_shum_isentr)
     endif

     if(ozone)   then
        ozone_isavg  = ozone_isavg   + mrdnl_avg(ozone_isentr)
        vozone_isavg = vozone_isavg  + mrdnl_avg(vozone_isentr)
     endif


     ipv_isavg       = ipv_isavg    + mrdnl_avg(abs_vor_isentr)
     ipv2_isavg      = ipv2_isavg   + mrdnl_avg(ipv*abs_vor_isentr)
     ipv_flux_isavg  = ipv_flux_isavg +                                       &
                       mrdnl_avg(v_isentr * abs_vor_isentr)
     ! temp. stuff added for tapio:
     v_theta_isavg   = v_theta_isavg +                                        &
          mrdnl_avg(vrtcl_gradient(v_isentr,theta_coord))
     v_isavg_nw      = v_isavg_nw   + mrdnl_avg(v_isentr)
                                                                               
     pv_isavg        = pv_isavg     + mrdnl_avg(v_isentr*p_isentr)
     pv_theta_isavg  = pv_theta_isavg                                         &
          + mrdnl_avg(p_isentr*vrtcl_gradient(v_isentr,theta_coord))
     rhov_isavg      = rhov_isavg   + mrdnl_avg(v_isentr*dens_isentr)

     mont_x_isavg    = mont_x_isavg + mrdnl_avg(mont_x_isentr)
     rho_mont_x_isavg= rho_mont_x_isavg + mrdnl_avg(rho_mont_x_isentr)

     mont_isavg      = mont_isavg   + mrdnl_avg(dens_isentr * mont_isentr)
     mont_var_isavg  = mont_var_isavg+ mrdnl_avg(dens_isentr * mont_isentr**2)


  end subroutine theta_averages

! ----------------------------------------------------------------------------

  subroutine average_fields(dof)

!----------------------------- input arguments --------------------------------

    real, intent(in) :: dof ! degrees of freedom

!----------------------------- local variables  -------------------------------
    
    integer j

!----------------------------- executable code  -------------------------------
    
     ! divide averages by number of degrees of freedom in estimates
    ps_grid_avg          = ps_grid_avg / dof

    ! divide sigma-averaged fields by mean sfc pressure and dof's

    u_avg                = divide_by_psfc(u_avg/dof,              ps_grid_avg)
    u_var_avg            = divide_by_psfc(u_var_avg/dof,          ps_grid_avg)
    v_avg                = divide_by_psfc(v_avg/dof,              ps_grid_avg)
    vort_avg             = divide_by_psfc(vort_avg/dof,           ps_grid_avg)
    v_var_avg            = divide_by_psfc(v_var_avg/dof,          ps_grid_avg)
    w_var_avg            = divide_by_psfc(w_var_avg/dof,          ps_grid_avg)
    w_cube_avg           = divide_by_psfc(w_cube_avg/dof,          ps_grid_avg)

    if(moisture) then
       shum_avg                 = divide_by_psfc(shum_avg/dof,               ps_grid_avg)
       shum_var_avg             = divide_by_psfc(shum_var_avg/dof,           ps_grid_avg)
       sat_shum_avg             = divide_by_psfc(sat_shum_avg/dof,           ps_grid_avg)
       sat_shum_var_avg         = divide_by_psfc(sat_shum_var_avg/dof,       ps_grid_avg)
       vshum_avg                = divide_by_psfc(vshum_avg/dof,              ps_grid_avg)
       wshum_avg                = divide_by_psfc(wshum_avg/dof,              ps_grid_avg)
       temp_shum_avg            = divide_by_psfc(temp_shum_avg/dof,           ps_grid_avg)
       rhum_avg                 = divide_by_psfc(rhum_avg/dof,               ps_grid_avg)
       theta_e_avg              = divide_by_psfc(theta_e_avg/dof,            ps_grid_avg)
       theta_e_sat_avg          = divide_by_psfc(theta_e_sat_avg/dof,        ps_grid_avg)
       theta_e_var_avg          = divide_by_psfc(theta_e_var_avg/dof,        ps_grid_avg) 
       mrdnl_theta_e_flux_avg   = divide_by_psfc(mrdnl_theta_e_flux_avg/dof, ps_grid_avg)
       vrtcl_theta_e_flux_avg   = divide_by_psfc(vrtcl_theta_e_flux_avg/dof, ps_grid_avg)

       do j=1,num_bin
         rhum_pdf(:,:,j)        = divide_by_psfc(rhum_pdf(:,:,j)/dof,        ps_grid_avg)
       end do

    endif

    if (moist_tend) then
       dt_qg_cond_avg  = divide_by_psfc(dt_qg_cond_avg/dof,   ps_grid_avg)
       dt_qg_conv_avg  = divide_by_psfc(dt_qg_conv_avg/dof,   ps_grid_avg)
       dt_qg_cond_freq = divide_by_psfc(dt_qg_cond_freq/dof,   ps_grid_avg)
       dt_qg_conv_freq = divide_by_psfc(dt_qg_conv_freq/dof,   ps_grid_avg)
    endif

    if (precip) then
       precip_cond_avg  = precip_cond_avg/dof
       precip_conv_avg  = precip_conv_avg/dof

       precip_intensity      = precip_intensity/wet_days
       precip_cond_intensity = precip_cond_intensity/wet_days_cond
       precip_conv_intensity = precip_conv_intensity/wet_days_conv
    endif

    if(ozone) then
       ozone_avg         = divide_by_psfc(ozone_avg/dof,          ps_grid_avg)
       vozone_avg        = divide_by_psfc(vozone_avg/dof,         ps_grid_avg)
    endif


    uvcos_avg            = divide_by_psfc(uvcos_avg/dof,          ps_grid_avg)
    wu_avg               = divide_by_psfc(wu_avg/dof,             ps_grid_avg)
    z_avg                = divide_by_psfc(z_avg/dof,              ps_grid_avg)
    temp_avg             = divide_by_psfc(temp_avg/dof,           ps_grid_avg)
    virtual_temp_avg     = divide_by_psfc(virtual_temp_avg/dof,   ps_grid_avg)
    temp_var_avg         = divide_by_psfc(temp_var_avg/dof,       ps_grid_avg) 
    pot_temp_avg         = divide_by_psfc(pot_temp_avg/dof,       ps_grid_avg)
    pot_temp_var_avg     = divide_by_psfc(pot_temp_var_avg/dof,   ps_grid_avg) 
    vcos_avg             = divide_by_psfc(vcos_avg/dof,           ps_grid_avg) 
    w_avg                = divide_by_psfc(w_avg/dof,              ps_grid_avg) 
    w_up_avg             = divide_by_psfc(w_up_avg/dof,           ps_grid_avg)
    w_up_freq            = w_up_freq / dof

    mrdnl_heatflux_avg   = divide_by_psfc(mrdnl_heatflux_avg/dof, ps_grid_avg)
    vrtcl_heatflux_avg   = divide_by_psfc(vrtcl_heatflux_avg/dof, ps_grid_avg)
    mrdnl_tempflux_avg   = divide_by_psfc(mrdnl_tempflux_avg/dof, ps_grid_avg)
    vrtcl_tempflux_avg   = divide_by_psfc(vrtcl_tempflux_avg/dof, ps_grid_avg)
    mrdnl_z_flux_avg     = divide_by_psfc(mrdnl_z_flux_avg/dof,   ps_grid_avg)
    vrtcl_z_flux_avg     = divide_by_psfc(vrtcl_z_flux_avg/dof,   ps_grid_avg)
    svol_avg             = divide_by_psfc(svol_avg/dof,           ps_grid_avg)
    hor_sfctn_avg        = hor_sfctn_avg/dof
    hor_sfctn_var_avg    = hor_sfctn_var_avg/dof
    hor_sfctn_avg        = radius**2 * hor_sfctn_avg
    hor_sfctn_var_avg    = radius**4 * hor_sfctn_var_avg

    sfctn_avg            = sfctn_avg            / dof
    sfctn_isavg          = sfctn_isavg          / dof

    if(moisture) then
       shum_isavg            = shum_isavg              / dof
       rhum_isavg            = rhum_isavg              / dof
       sat_shum_isavg        = sat_shum_isavg          / dof
       vshum_isavg           = vshum_isavg             / dof
       vsat_shum_isavg       = vsat_shum_isavg         / dof
    endif

    if(ozone) then
       ozone_isavg       = ozone_isavg          / dof
       vozone_isavg      = vozone_isavg         / dof
    endif

    p_full_avg           = p_full_avg           / dof
    freq_stat_unstable   = freq_stat_unstable   / dof
    sfc_frequency        = sfc_frequency        / dof / float(num_lon)
    p_isavg              = p_isavg              / dof
    p_var_isavg          = p_var_isavg          / dof
    exner_isavg          = exner_isavg          / dof
    pexner_isavg         = pexner_isavg         / dof
    dens_isavg           = dens_isavg           / dof
    v_isavg              = v_isavg              / dof
    u_isavg              = u_isavg              / dof
    v2_isavg             = v2_isavg             / dof
    u2_isavg             = u2_isavg             / dof
    ipv_isavg            = ipv_isavg            / dof
    ipv2_isavg           = ipv2_isavg           / dof
    ipv_flux_isavg       = ipv_flux_isavg       / dof
    spv_flux_y_avg       = spv_flux_y_avg       / dof
    geos_heat_flux_sfc   = geos_heat_flux_sfc   / dof
    v_theta_isavg        = v_theta_isavg        / dof
    v_isavg_nw           = v_isavg_nw           / dof
    pv_isavg             = pv_isavg             / dof
    pv_theta_isavg       = pv_theta_isavg       / dof
    rhov_isavg           = rhov_isavg           / dof

    ! these are multiplied by psfc in the vertical averaging routine
    u_baro_avg           = u_baro_avg / ( dof * ps_grid_avg )
    v_baro_avg           = v_baro_avg / ( dof * ps_grid_avg )

    u_baro_var_avg       = u_baro_var_avg / (dof * ps_grid_avg )
    v_baro_var_avg       = v_baro_var_avg / (dof * ps_grid_avg )

    do i=1, num_lat
       ucos_zon_spec_avg(i,:) = ucos_zon_spec_avg(i,:) / (dof * ps_grid_avg(i))
       vcos_zon_spec_avg(i,:) = vcos_zon_spec_avg(i,:) / (dof * ps_grid_avg(i))
       if(moisture)                                                           &
            shum_zon_spec_avg(i,:) = shum_zon_spec_avg(i,:) /                 &
                                     (dof * ps_grid_avg(i))
       if(ozone)                                                              &
            ozone_zon_spec_avg(i,:) = ozone_zon_spec_avg(i,:) /               &
                                     (dof * ps_grid_avg(i))
    enddo

    mont_x_isavg         = mont_x_isavg         / dof
    mont_isavg           = mont_isavg           / dof
    mont_var_isavg       = mont_var_isavg       / dof
    rho_mont_x_isavg     = rho_mont_x_isavg     / dof

    bt_sfctn_avg         = radius**2 * bt_sfctn_avg / dof
    bt_sfctn_var_avg     = radius**4 * bt_sfctn_var_avg / dof

    bc_sfctn_spec        = bc_sfctn_spec / dof
    bt_sfctn_spec        = bt_sfctn_spec / dof

    pot_temp_spec        = pot_temp_spec / dof
    conv_spec            = conv_spec     / dof
    if (moisture) then  
      shum_spec          = shum_spec     / dof
      sat_shum_spec      = sat_shum_spec     / dof
      rhum_spec          = rhum_spec     / dof
      theta_e_spec       = theta_e_spec  / dof
    end if
    if (ozone)                                                              &
      ozone_spec   = ozone_spec    / dof


   ! normalization factors for spherepack:
    bc_sfctn_spec(1,:) = 0.25 * radius**4 * bc_sfctn_spec(1,:)
    bc_sfctn_spec(2:num_fourier+1,:) =                                        &
         0.25 * radius**4 * bc_sfctn_spec(2:num_fourier+1,:) 

    bt_sfctn_spec(1,:) = 0.25 * radius**4 * bt_sfctn_spec(1,:)
    bt_sfctn_spec(2:num_fourier+1,:) =                                        &
         0.25 * radius**4 * bt_sfctn_spec(2:num_fourier+1,:)


    pot_temp_spec(1,:,:) = 0.25 * pot_temp_spec(1,:,:)
    pot_temp_spec(2:num_fourier+1,:,:) =                                      &
         0.25 * pot_temp_spec(2:num_fourier+1,:,:)

    conv_spec(1,:,:) = 0.25 * conv_spec(1,:,:)
    conv_spec(2:num_fourier+1,:,:) =                                          &
         0.25 * conv_spec(2:num_fourier+1,:,:)

    if (moisture) then
 
       shum_spec(1,:,:) = 0.25 * shum_spec(1,:,:)
       shum_spec(2:num_fourier+1,:,:) =                                       &
            0.25 * shum_spec(2:num_fourier+1,:,:)

       sat_shum_spec(1,:,:) = 0.25 * sat_shum_spec(1,:,:)
       sat_shum_spec(2:num_fourier+1,:,:) =                                   &
            0.25 * sat_shum_spec(2:num_fourier+1,:,:)

       rhum_spec(1,:,:) = 0.25 * rhum_spec(1,:,:)
       rhum_spec(2:num_fourier+1,:,:) =                                       &
            0.25 * rhum_spec(2:num_fourier+1,:,:)

       theta_e_spec(1,:,:) = 0.25 * theta_e_spec(1,:,:)
       theta_e_spec(2:num_fourier+1,:,:) =                                    &
            0.25 * theta_e_spec(2:num_fourier+1,:,:)

    endif

    if(ozone) then
       ozone_spec(1,:,:) = 0.25 * ozone_spec(1,:,:)
       ozone_spec(2:num_fourier+1,:,:) =                                      &
            0.25 * ozone_spec(2:num_fourier+1,:,:)
    endif


    theta_s_mont_x       = theta_s_mont_x       / dof

  end subroutine average_fields

! ----------------------------------------------------------------------------
  
  subroutine eddy_fields

!----------------------------- local variables  -------------------------------

    integer  :: k

    real, dimension(num_lat, num_lev) ::                                      &
         vrtcl_eddy_tempflux_sig,&! derivative of meridional temp. flux WRT sig
         wu_eddy_sig,         &   ! derviative of \overline{w'u'} WRT sig
         mean_conv                ! conversion of PE to KE
    

!----------------------------- executable code  -------------------------------



! temp. stuff for tapio:
    pv_eddy_isavg        = pv_isavg - v_isavg_nw * p_isavg
    pv_theta_eddy_isavg  = pv_theta_isavg - p_isavg * v_theta_isavg
    rhov_eddy_isavg      = rhov_isavg - v_isavg_nw * dens_isavg
!
    p_var_isavg          = p_var_isavg - p_isavg**2

    temp_var_avg         = temp_var_avg - temp_avg**2
    pot_temp_var_avg     = pot_temp_var_avg - pot_temp_avg**2
    v_var_avg            = v_var_avg - v_avg**2
    u_var_avg            = u_var_avg - u_avg**2
    w_var_avg            = w_var_avg - w_avg**2
    w_cube_avg           = w_cube_avg -3.0*w_var_avg*w_avg - w_avg**3


    u_baro_var_avg       = u_baro_var_avg - u_baro_avg**2
    v_baro_var_avg       = v_baro_var_avg - v_baro_avg**2

    hor_sfctn_var_avg    = hor_sfctn_var_avg - hor_sfctn_avg**2

    bt_sfctn_var_avg     = bt_sfctn_var_avg - bt_sfctn_avg**2

    uvcos_eddy           = uvcos_avg - u_avg * vcos_avg
    wu_eddy              = wu_avg - w_avg * u_avg

    mrdnl_eddy_heatflux  = mrdnl_heatflux_avg - vcos_avg * pot_temp_avg
    vrtcl_eddy_heatflux  = vrtcl_heatflux_avg - w_avg * pot_temp_avg

    mrdnl_eddy_z_flux    = mrdnl_z_flux_avg - vcos_avg * z_avg
    vrtcl_eddy_z_flux    = vrtcl_z_flux_avg - w_avg * z_avg

    if (moisture) then
      shum_var_avg             = shum_var_avg - shum_avg**2
      sat_shum_var_avg         = sat_shum_var_avg - sat_shum_avg**2
      theta_e_var_avg          = theta_e_var_avg - theta_e_avg**2
      mrdnl_eddy_theta_e_flux  = mrdnl_theta_e_flux_avg - vcos_avg * theta_e_avg
      vrtcl_eddy_theta_e_flux  = vrtcl_theta_e_flux_avg - w_avg * theta_e_avg
      temp_shum_avg            = temp_shum_avg - temp_avg * shum_avg
    end if

    mrdnl_eddy_tempflux  = mrdnl_tempflux_avg - vcos_avg * temp_avg
    vrtcl_eddy_tempflux  = vrtcl_tempflux_avg - w_avg * temp_avg
    
    ! derivative of vertical eddy flux WRT sigma:
    vrtcl_eddy_tempflux_sig = divide_by_psfc(                                 &
         vrtcl_gradient(vrtcl_eddy_tempflux, sigma), ps_grid_avg)

    ! derivative of \overline{w'u'} WRT sigma:
    wu_eddy_sig = divide_by_psfc(                                             &
         vrtcl_gradient(wu_eddy, sigma), ps_grid_avg)

    ! divergence of eddy temperature flux
    ! mrdnl_eddy_tempflux, uvcos_eddy, and v_var_avg are 
    ! \frac{\overline{ps cos(lat) v' X'}}{\overline{ps}}
    ! We want to take the derivative of \overline{ps cos(lat) v' X'}, so
    ! need to multiply by \overline{ps}

    eddy_temp_div      =                                                      &
         - (mrdnl_gradient(mrdnl_eddy_tempflux / cos_lat_yz, deg_lat)) -  &
         vrtcl_eddy_tempflux_sig

  end subroutine eddy_fields

! ----------------------------------------------------------------------------

  subroutine read_namelist

    integer :: unit

    unit = 20
    rewind(unit)
    open(unit, file='input.nml')
    read(unit, nml=main_list)
    rewind(unit)
    read(unit, nml=filename_list)

    close(unit)


  end subroutine read_namelist

end program offlinediag
  

