module output_mod

  use vars
  use typesizes, only: bytesizesOK
  use netcdf
  use constants_and_switches_mod, only: gspval

implicit none
private

public :: create_output_file, write_variables, close_output_file

  integer, parameter ::                                                       &
       nvars = 1002,       &  ! number of output fields (max)
       theta_switch = 1,   &  ! theta variable switch
       sigma_switch = 2,   &  ! sigma variable switch
       surface_switch = 3, &  ! surface variable switch
       zon_wave_switch = 4,&  ! zonal wavenumber variable switch
       spectra_switch = 5, &  ! spectral variable switch
       pdf_switch = 6,     &  ! pdf (lat, lev, bin)
       spectra_switch_3d = 7  ! 3d spectral (wn, wn, height)

  integer ::                                                                  &
       ncFileID,           &  ! file ID

       latDimID,           &  ! latitude file ID
       sigmaDimID,         &  ! sigma dimension ID
       thetaDimID,         &  ! theta dimension ID
       zon_wavenDimID,     &  ! zonal wavenumber dimension ID 
       legendreDimID,      &  ! legendre dimension ID
       binDimID,           &  ! pdf bin dimension ID
       timeDimID,          &  ! time dimension ID
       timeVarID,          &  ! time variable ID
       latVarID,           &  ! latitude variable ID
       sigmaVarID,         &  ! sigma variable ID
       thetaVarID,         &  ! theta variable ID
       zon_wavenVarID,     &  ! zonal wavenumber variable ID
       legendreVarID,      &  ! legendre variable ID
       binVarID,           &  ! pdf bin variable ID
       statVarID(nvars)       ! array of field variable IDs

contains
! #############################################################################

  subroutine create_output_file(                                              &
                             num_lat,                         num_lev )

! creates output file and axis variables
    
!----------------------------- input arguments --------------------------------
    
    integer, intent(in) :: num_lat, num_lev
    
!----------------------------- local variables --------------------------------

!----------------------------- executable code --------------------------------

    if(.not. byteSizesOK()) then
       stop 'open_and_initialize: Compiler does not appear to support required kinds of variables'
    end if

    ! -----------   create netCDF file and define variables  -------------
    ! create netcdf file
    call check(nf90_create(path = trim(OutputFileName),                       &
         cmode = nf90_share, ncid = ncFileID))

    ! global attributes
    call check(nf90_put_att(ncFileID, nf90_global, "title",                   &
         "OfflineDiag Analyses") )
       
       ! global attributes
    call check(nf90_put_att(ncFileID, nf90_global, "title",                   &
         "OfflineDiag Analyses") )


    ! Define axes
    ! -----------
    ! latitude
    call check(NF90_DEF_DIM(ncFileID, "lat", num_lat, latDimID) )
    call check(nf90_def_var(ncFileID, "lat", nf90_double, latDimID,           &
         latVarID) )  
    call check(nf90_put_att(ncFileID, latVarID,                               &
         "long_name", "latitude") )
    call check(nf90_put_att(ncFileID, latVarID,                               &
         "units", "degrees_north") )
       
    ! sigma level
    call check(NF90_DEF_DIM(ncFileID, "sigma", num_lev, sigmaDimID) )
    call check(nf90_def_var(ncFileID, "sigma", nf90_double, sigmaDimID,       &
         sigmaVarID))
    call check(nf90_put_att(ncFileID, sigmaVarID,                             &
         "long_name", "sigma level") )
    call check(nf90_put_att(ncFileID, sigmaVarID,                             &
         "units", "sigma_level") )
    
    ! potential temperature
    call check(NF90_DEF_DIM(ncFileID, "theta", MaxIsentrLev, thetaDimID) )
    call check(nf90_def_var(ncFileID, "theta", nf90_double, thetaDimID,       &
         thetaVarID))
    call check(nf90_put_att(ncFileID, thetaVarID,                             &
         "long_name", "potential temperature") )
    call check(nf90_put_att(ncFileID, thetaVarID,                             &
         "units", "deg_K") )
    
    ! zonal wavenumber
    call check(nf90_def_dim(ncFileID, "zon_waven", num_fourier+1,             &
         zon_wavenDimID) )
    call check(nf90_def_var(ncFileID, "zon_waven", nf90_double,               &
         zon_wavenDimID, zon_wavenVarID) )
    call check(nf90_put_att(ncFileID, zon_wavenVarID,                         &
         "long_name", "zonal wavenumber") )
    call check(nf90_put_att(ncFileID, zon_wavenVarID,                         &
         "units", "none") )
       
    ! zonal wavenumber
    call check(nf90_def_dim(ncFileID, "legendre", num_fourier+1,              &
         legendreDimID) )
    call check(nf90_def_var(ncFileID, "legendre", nf90_double,                &
         legendreDimID, legendreVarID) )
    call check(nf90_put_att(ncFileID, legendreVarID,                          &
         "long_name", "legendre wavenumber") )
    call check(nf90_put_att(ncFileID, legendreVarID,                          &
         "units", "none") )

    ! pdf bin for relative humidity
    call check(NF90_DEF_DIM(ncFileID, "rhum_bin", num_bin, binDimID) )
    call check(nf90_def_var(ncFileID, "rhum_bin", nf90_double, binDimID,      &
         binVarID))
    call check(nf90_put_att(ncFileID, binVarID,                               &
         "long_name", "relative humidity bin") )
    call check(nf90_put_att(ncFileID, binVarID,                               &
         "units", "none") )
       
    ! time
    call check(NF90_DEF_DIM(ncFileID, "times", 1, timeDimID) )
    call check(nf90_def_var(ncFileID, "times", nf90_double, timeDimID,        &
         timeVarID) )  
    call check(nf90_put_att(ncFileID, timeVarID,                              &
         "long_name", "number of instants") )
       
    call initialize_variables
       
     end subroutine create_output_file
    
! #############################################################################

     subroutine initialize_variables

    ! Define variables
    ! ----------------

    call variable_init("is_streamfctn",                                       &
         "Isentropic Streamfunction",                                         &
         "kg/s",                                                              &
         theta_switch,                                                        &
         statVarID(1))

    call variable_init("sfc_frequency",                                       &
         "Frequency Distribution of Sfc Temperatures",                        &
         "1",                                                                 &
         theta_switch,                                                        &
         statVarID(2))

    call variable_init("d_dy_pot_temp",                                       &
         "Meridional Potential Temperature Gradient",                         &
         "K/m",                                                               &
         sigma_switch,                                                        &
         statVarID(3))

    call variable_init("d_dp_pot_temp",                                       &
         "Vrtcl Potential Temperature Gradient (press.)",                     &
         "K/Pa",                                                              &
         sigma_switch,                                                        &
         statVarID(4))

    call variable_init("mrdnl_heatflux",                                      &
         "Meridional Eddy Heatflux",                                          &
         "K.m/s",                                                             &
         sigma_switch,                                                        &
         statVarID(5))
    
    call variable_init("u",                                                   &
         "Zonal wind",                                                        &
         "m/s",                                                               &
         sigma_switch,                                                        &
         statVarID(6))

    call variable_init("temp",                                                &
         "Temperature",                                                       &
         "K",                                                                 &
         sigma_switch,                                                        &
         statVarID(7))

    call variable_init("pot_temp",                                            &
         "Potential Temperature",                                             &
         "K",                                                                 &
         sigma_switch,                                                        &
         statVarID(8))

    call variable_init("buoyancy_freq",                                       &
         "Buoyancy Frequency",                                                &
         "1/s",                                                               &
         sigma_switch,                                                        &
         statVarID(9))
    call check(nf90_put_att(ncFileID, statVarID(9),                           &
         "_FillValue", gspval ) )

    call variable_init("ucos_zon_spec",                                       &
         "Spectrum of Zonal Kinetic Energy",                                  &
         "m.m/s/s",                                                           &
         zon_wave_switch,                                                     &
         statVarID(10))

    call variable_init("vcos_zon_spec",                                       &
         "Spectrum of Meridional Kinetic Energy",                             &
         "m.m/s/s",                                                           &
         zon_wave_switch,                                                     &
         statVarID(11))

    call variable_init("p_isentr",                                            &
         "Pressure on Isentropic Levels",                                     &
         "Pa",                                                                &
         theta_switch,                                                        &
         statVarID(12))

    call variable_init("p_full",                                              &
         "Pressure on Sigma Levels",                                          &
         "Pa",                                                                &
         sigma_switch,                                                        &
         statVarID(13))

    call variable_init("temp_var",                                            &
         "Temperature Variance",                                              &
         "K.K",                                                               &
         sigma_switch,                                                        &
         statVarID(14))

    call variable_init("streamfctn",                                          &
         "Eulerian mass transport streamfunction",                            &
         "kg/s",                                                              &
         sigma_switch,                                                        &
         statVarID(15))
    
    call variable_init("ipv_flux",                                            &
         "Meridional flux of IPV",                                            &
         "m.m.m.deg_K/s/s/kg",                                                &
         theta_switch,                                                        &
         statVarID(16))

    call variable_init("dens_isavg",                                          &
         "Density in isentropic coords",                                      &
         "kg/m/m/deg_K",                                                      &
         theta_switch,                                                        &
         statVarID(17))

    call variable_init("ipv_isavg",                                           &
         "Density-weighted average of IPV",                                   &
         "1/s",                                                               &
         theta_switch,                                                        &
         statVarID(18))

    call variable_init("v_isavg",                                             &
         "Mass-weighted average of v",                                        &
         "kg/m/deg_K/s",                                                      &
         theta_switch,                                                        &
         statVarID(19))

    call variable_init("ipv2_isavg",                                          &
         "Isentropic-mean squared PV",                                        &
         "m.m.deg_K/kg/s/s",                                                  &
         theta_switch,                                                        &
         statVarID(20))

    call variable_init("pot_temp_var",                                        &
         "Potential Temperature Variance",                                    &
         "K.K",                                                               &
         sigma_switch,                                                        &
         statVarID(21))

    call variable_init("v",                                                   &
         "Meridional wind",                                                   &
         "m/s",                                                               &
         sigma_switch,                                                        &
         statVarID(22))
    
    call variable_init("v_var",                                               &
         "Meridional wind variance",                                          &
         "m.m/s/s",                                                           &
         sigma_switch,                                                        &
         statVarID(23))

    call variable_init("u_var",                                               &
         "Zonal wind variance",                                               &
         "m.m/s/s",                                                           &
         sigma_switch,                                                        &
         statVarID(24))

    call variable_init("uvcos_eddy",                                          &
         "Meridional eddy flux of zonal momentum",                            &
         "m.m/s/s",                                                           &
         sigma_switch,                                                        &
         statVarID(25))

    call variable_init("v2_isavg",                                            &
         "Isentropic-mean squared v",                                         &
         "kg/s/s/deg_K",                                                      &
         theta_switch,                                                        &
         statVarID(26))

    call variable_init("spv_flux_y_avg",                                      &
         "Integ. y-component of surface potential vorticity flux",            &
         "m.deg_K/s/s",                                                       &
         surface_switch,                                                      &
         statVarID(27))

    call variable_init("exner",                                               &
         "Exner function",                                                    &
         "m.m/s/s/K",                                                         &
         theta_switch,                                                        &
         statVarID(28))

    call variable_init("freq_stat_unstable",                                  &
         "Frequency of statically unstable conditions",                       &
         "1",                                                                 &
         sigma_switch,                                                        &
         statVarID(29))

    call variable_init("mrdnl_tempflux",                                      &
         "Meridional temperature flux",                                       &
         "K.m/s",                                                             &
         sigma_switch,                                                        &
         statVarID(30))

    call variable_init("svol",                                                &
         "Specific volume",                                                   &
         "m.m.m/kg",                                                          &
         sigma_switch,                                                        &
         statVarID(31))

    call variable_init("geos_heat_flux_sfc",                                  &
         "Geostrophic surface pot. temp. flux (times f)",                     &
         "m.deg_K/s/s",                                                       &
         surface_switch,                                                      &
         statVarID(32))

    call variable_init("res_circ1",                                           &
         " TEM Residual Circulation",                                         &
         "kg/s",                                                              &
         sigma_switch,                                                        &
         statVarID(33))

    call variable_init("res_circ2",                                           &
         "Modified TEM Residual Circulation",                                 &
         "kg/s",                                                              &
         sigma_switch,                                                        &
         statVarID(34))

    call variable_init("vrtcl_heatflux",                                      &
         "Vertical Heatflux",                                                 &
         "K.m/s",                                                             &
         sigma_switch,                                                        &
         statVarID(35))

    call variable_init("vrtcl_eddy_heatflux",                                 &
         "Vertical Eddy Heatflux",                                            &
         "K.m/s",                                                             &
         sigma_switch,                                                        &
         statVarID(36))

    call variable_init("mrdnl_eddy_tempflux",                                 &
         "Meridional eddy temperature flux",                                  &
         "K.m/s",                                                             &
         sigma_switch,                                                        &
         statVarID(37))

    call variable_init("vrtcl_eddy_tempflux",                                 &
         "Vertical eddy temperature flux",                                    &
         "K.Pa/s",                                                            &
         sigma_switch,                                                        &
         statVarID(38))

    call variable_init("eddy_temp_flux_div",                                  &
         "divergence of eddy temperature flux",           &
         "K/s",                                                               &
         sigma_switch,                                                        &
         statVarID(39))

    call variable_init("ps_grid_avg",                                         &
         "Surface pressure",                                                  &
         "Pa",                                                                &
         surface_switch,                                                      &
         statVarID(40))

    call variable_init("w_avg",                                               &
         "vertical wind; w = \dot\sigma",                                     &
         "1/s",                                                               &
         sigma_switch,                                                        &
         statVarID(41))

    call variable_init("vrtcl_tempflux",                                      &
         "Vertical temperature flux",                                         &
         "K.m/s",                                                             &
         sigma_switch,                                                        &
         statVarID(42))

    call variable_init("wu_eddy",                                             &
         "Vertical eddy momentum flux",                                       &
         "K.m/s",                                                             &
         sigma_switch,                                                        &
         statVarID(43))

    call variable_init("pv_eddy",                                             &
         "overline{p'v'}",                                                    &
         "Pa.m/s",                                                            &
         theta_switch,                                                        &
         statVarID(44))

    call variable_init("pv_theta_eddy",                                       &
         "overline{p'v_theta'}",                                              &
         "Pa.m/s/K",                                                          &
         theta_switch,                                                        &
         statVarID(45))

    call variable_init("eddy_mass_flux",                                      &
         "overline{rho'.v'}",                                                 &
         "kg/m/K/s",                                                          &
         theta_switch,                                                        &
         statVarID(46))

    call variable_init("p_var_isavg",                                         &
         "variance of pressure",                                              &
         "Pa.Pa",                                                             &
         theta_switch,                                                        &
         statVarID(47))

    call variable_init("u2_isavg",                                            &
         "Isentropic-mean squared u",                                         &
         "kg/s/s/deg_K",                                                      &
         theta_switch,                                                        &
         statVarID(48))

    call variable_init("u_isavg",                                             &
         "Mass-weighted average of u",                                        &
         "kg/m/deg_K/s",                                                      &
         theta_switch,                                                        &
         statVarID(49))

    call variable_init("u_baro_var",                                          &
         "variance of barotropic zonal wind",                                 &
         "m.m/s/s",                                                           &
         surface_switch,                                                      &
         statVarID(50))

    call variable_init("v_baro_var",                                          &
         "variance of barotropic meridional wind",                            &
         "m.m/s/s",                                                           &
         surface_switch,                                                      &
         statVarID(51))

    call variable_init("pexner",                                              &
         "p*exner",                                                           &
         "none",                                                              &
         theta_switch,                                                        &
         statVarID(52))

    call variable_init("fv_g",                                                &
         "f times geostrophic velocity",                                      &
         "m/s/s",                                                             &
         theta_switch,                                                        &
         statVarID(53))

    call variable_init("fV_g",                                                &
         "f times density-weighted geostrophic velocity",                     &
         "kg.m/s/s/K",                                                        &
         theta_switch,                                                        &
         statVarID(54))

    call variable_init("spec_baroclinic",                                     &
         "spectral coefficients of the baroclinic streamfunction",            &
         "none",                                                              &
         spectra_switch,                                                      &
         statVarID(55))

    call variable_init("spec_barotropic",                                     &
         "spectral coefficients of the barotropic streamfunction",            &
         "none",                                                              &
         spectra_switch,                                                      &
         statVarID(56))

    call variable_init("theta_s_mont_x",                                      &
         "theta_s**2 * montgomery_x",                                         &
         "K.K.m/s/s",                                                         &
         surface_switch,                                                      &
         statVarID(57))
    
    call variable_init("bt_streamfctn_avg",                                   &
         "barotropic streamfctn",                                             &
         "m.m/s",                                                             &
         surface_switch,                                                      &
         statVarID(58))

    call variable_init("bt_streamfctn_var_avg",                               &
         "barotropic streamfctn variance",                                    &
         "m.m.m.m/s/s",                                                       &
         surface_switch,                                                      &
         statVarID(59))

    call variable_init("mont_isavg",                                          &
         "Montgomery streamfunction in isentropic coords",                    &
         "m.m/s/s",                                                           &
         theta_switch,                                                        &
         statVarID(60))

    call variable_init("mont_var_isavg",                                      &
         "variance Montgomery streamfunction in isentropic coords",           &
         "m.m.m.m/s/s/s/s",                                                   &
         theta_switch,                                                        &
         statVarID(61))

    call variable_init("pot_temp_spec",                                       &
         "potential temperature spectrum",                                    &
         "none",                                                              &
         spectra_switch_3d,                                                   &
         statVarID(62))

    call variable_init("conv_spec",                                           &
         "spectrum of pe -> ke conversion",                                   &
         "none",                                                              &
         spectra_switch_3d,                                                   &
         statVarID(63))

    call variable_init("hor_sfctn_avg",                                       &
         "horizontal streamfunction",                                         &
         "m.m/s",                                                             &
         sigma_switch,                                                        &
         statVarID(64))
    
    call variable_init("hor_sfctn_var_avg",                                   &
         "horizontal streamfunction variance",                                &
         "m.m.m.m/s/s",                                                       &
         sigma_switch,                                                        &
         statVarID(65))
  
    if(moisture) then

       call variable_init("shum_avg",                                         &
            "Specific humidity",                                              &
            "kg/kg",                                                          &
            sigma_switch,                                                     &
            statVarID(66))
       
       call variable_init("rhum_avg",                                         &
            "Relative humidity",                                              &
            "none",                                                           &
            sigma_switch,                                                     &
            statVarID(67))
       
       call variable_init("shum_isavg",                                       &
            "Specific humidity",                                              &
            "kg/kg",                                                          &
            theta_switch,                                                     &
            statVarID(68))

       call variable_init("sat_shum_isavg",                                   &
            "Saturation specific humidity",                                   &
            "kg/kg",                                                          &
            theta_switch,                                                     &
            statVarID(69))

       call variable_init("vshum_avg",                                        &
            "Meridional flux of Specific humidity",                           &
            "kg.m/kg/s",                                                      &
            sigma_switch,                                                     &
            statVarID(70))


       call variable_init("vshum_isavg",                                      &
            "Meridional flux of Specific humidity",                           &
            "kg.m/kg/s",                                                      &
            theta_switch,                                                     &
            statVarID(71))

       call variable_init("shum_zon_spec",                                    &
            "Zonal spectrum of specific humidity",                            &
            "kg/kg",                                                          &
            zon_wave_switch,                                                  &
            statVarID(72))

       call variable_init("shum_spec",                                        &
            "spectrum of specific humidity",                                  &
            "none",                                                           &
            spectra_switch_3d,                                                &
            statVarID(73))

       call variable_init("mrdnl_eddy_theta_e_flux",                          &
            "Meridional Eddy Equivalent Potential Temperature Flux",          &
            "K.m/s",                                                          &
            sigma_switch,                                                     &
            statVarID(74))
   
       call variable_init("theta_e",                                          &
            "Equivalent Potential Temperature",                               &
            "K",                                                              &
            sigma_switch,                                                     &
            statVarID(75))
   
       call variable_init("theta_e_var",                                      &
            "Equivalent Potential Temperature Variance",                      &
            "K.K",                                                            &
            sigma_switch,                                                     &
            statVarID(76))
   
       call variable_init("vrtcl_theta_e_flux",                               &
            "Vertical equiv. pot. temp flux",                                 &
            "K.m/s",                                                          &
            sigma_switch,                                                     &
            statVarID(77))

       call variable_init("vrtcl_eddy_theta_e_flux",                          &
            "Vertical Eddy equiv. pot. temp flux",                            &
            "K.m/s",                                                          &
            sigma_switch,                                                     &
            statVarID(78))
   
   
       call variable_init("theta_e_spec",                                     &
            "equivalent potential temperature spectrum",                      &
            "none",                                                           &
            spectra_switch_3d,                                                &
            statVarID(79))

       call variable_init("theta_e_sat",                                      &
            "Saturated Equivalent Potential Temperature",                     &
            "K",                                                              &
            sigma_switch,                                                     &
            statVarID(80))

       call variable_init("d_dy_theta_e",                                        &
            "Meridional Equivalent Potential Temperature Gradient",              &
            "K/m",                                                               &
            sigma_switch,                                                        &
            statVarID(81))
   
       call variable_init("d_dp_theta_e",                                        &
            "Vrtcl Equivalent Potential Temperature Gradient (press.)",          &
            "K/Pa",                                                              &
            sigma_switch,                                                        &
            statVarID(82))

       call variable_init("shum_var",                                            &
            "specific humidity variance",                                        &
            "kg/kg kg/kg",                                                       &
            sigma_switch,                                                        &
            statVarID(83))

       call variable_init("sat_shum",                                            &
            "saturated specific humidity",                                       &
            "kg/kg",                                                             &
            sigma_switch,                                                        &
            statVarID(84))

       call variable_init("rhum_pdf",                                            &
            "pdf of relative humidity",                                          &
            "none",                                                              &
            pdf_switch,                                                          &
            statVarID(85))

       call variable_init("wshum_avg",                                        &
            "Vertical flux of Specific humidity",                             &
            "kg/kg/s",                                                        &
            sigma_switch,                                                     &
            statVarID(86))

       call variable_init("temp_shum_avg",                                    &
            "Eddy correlation of temperature and specific humidity",          &
            "K.kg/kg",                                                        &
            sigma_switch,                                                     &
            statVarID(87))


       call variable_init("sat_shum_var",                                     &
            "saturation specific humidity variance",                          &
            "kg/kg kg/kg",                                                    &
            sigma_switch,                                                     &
            statVarID(88))

       call variable_init("vsat_shum_isavg",                                  &
            "Meridional flux of saturation specific humidity",                &
            "kg.m/kg/s",                                                      &
            theta_switch,                                                     &
            statVarID(89))

       call variable_init("rhum_isavg",                                       &
            "Relative humidity",                                              &
            "none",                                                           &
            theta_switch,                                                     &
            statVarID(90))
   
    endif

    call variable_init("z",                                                &
         "Geopotential",                                                   &
         "m",                                                              &
         sigma_switch,                                                     &
         statVarID(91))

    call variable_init("mrdnl_z_flux",                                     &
         "Meridional z Flux",                                              &
         "m.m/s",                                                          &
         sigma_switch,                                                     &
         statVarID(92))

    call variable_init("mrdnl_eddy_z_flux",                                &
         "Meridional Eddy z Flux",                                         &
         "m.m/s",                                                          &
         sigma_switch,                                                     &
         statVarID(93))

    if(ozone) then
                                                                                                                                     
       call variable_init("ozone_avg",                                        &
            "ozone",                                                          &
            "kg/kg",                                                          &
            sigma_switch,                                                     &
            statVarID(94))
                                                                                                                                     
       call variable_init("ozone_isavg",                                      &
            "ozone",                                                          &
            "kg/kg",                                                          &
            theta_switch,                                                     &
            statVarID(95))
                                                                                                                                     
       call variable_init("vozone_avg",                                       &
            "Meridional flux of ozone",                                       &
            "kg.m/kg/s",                                                      &
            sigma_switch,                                                     &
            statVarID(96))
                                                                                                                                     
       call variable_init("vozone_isavg",                                     &
            "Meridional flux of ozone",                                       &
            "kg.m/kg/s",                                                      &
            theta_switch,                                                     &
            statVarID(97))
                                                                                                                                     
       call variable_init("ozone_zon_spec",                                   &
            "Zonal spectrum of ozone",                                        &
            "kg/kg",                                                          &
            zon_wave_switch,                                                  &
            statVarID(98))
                                                                                                                                     
       call variable_init("ozone_spec",                                       &
            "spectrum of ozone",                                              &
            "none",                                                           &
            spectra_switch_3d,                                                &
            statVarID(99))
                                                                                                                                     
    endif

    if(moisture) then

       call variable_init("sat_shum_spec",                                    &
            "spectrum of saturation specific humidity",                       &
            "none",                                                           &
            spectra_switch_3d,                                                &
            statVarID(100))

       call variable_init("rhum_spec",                                        &
            "spectrum of relative humidity",                                  &
            "none",                                                           &
            spectra_switch_3d,                                                &
            statVarID(101))

    endif

    if(moist_tend) then

       call variable_init("dt_qg_cond_avg",                                   &
            "Large scale cond specific humidity tendency",                    &
            "kg/kg/s",                                                        &
            sigma_switch,                                                     &
            statVarID(102))

       call variable_init("dt_qg_conv_avg",                                   &
            "Convection specific humidity tendency",                          &
            "kg/kg/s",                                                        &
            sigma_switch,                                                     &
            statVarID(103))

       call variable_init("dt_qg_cond_freq",                                  &
            "Large scale cond moisture tendency frequency",                   &
            "none",                                                           &
            sigma_switch,                                                     &
            statVarID(104))

       call variable_init("dt_qg_conv_freq",                                  &
            "Convection moisture tendency frequency",                         &
            "none",                                                           &
            sigma_switch,                                                     &
            statVarID(105))

    endif

    call variable_init("w_var",                                               &
         "vertical wind variance",                                            &
         "1/s/s",                                                             &
         sigma_switch,                                                        &
         statVarID(106))

    if(precip) then

       call variable_init("precip_cond_avg",                                  &
            "Large scale cond precipitation",                                 &
            "kg/m/m/s",                                                       &
            surface_switch,                                                   &
            statVarID(107))

       call variable_init("precip_conv_avg",                                  &
            "Convection precipitation",                                       &
            "kg/m/m/s",                                                       &
            surface_switch,                                                   &
            statVarID(108))

       call variable_init("precip_intensity",                                 &
            "Precipitation intensity",                                        &
            "kg/m/m/s",                                                       &
            surface_switch,                                                   &
            statVarID(109))

       call variable_init("precip_cond_intensity",                            &
            "Large scale cond precip intensity",                              &
            "kg/m/m/s",                                                       &
            surface_switch,                                                   &
            statVarID(110))

       call variable_init("precip_conv_intensity",                            &
            "Convection precip intensity",                                    &
            "kg/m/m/s",                                                       &
            surface_switch,                                                   &
            statVarID(111))

    endif

    call variable_init("vrtcl_z_flux",                                     &
         "Vertical z Flux",                                                &
         "m/s",                                                            &
         sigma_switch,                                                     &
         statVarID(112))

    call variable_init("vrtcl_eddy_z_flux",                                &
         "Vertical Eddy z Flux",                                           &
         "m/s",                                                            &
         sigma_switch,                                                     &
         statVarID(113))

    call variable_init("w_cube",                                              &
         "vertical wind eddy cube",                                           &
         "1/s/s/s",                                                           &
         sigma_switch,                                                        &
         statVarID(114))

    call variable_init("w_up_avg",                                            &
         "upward vertical wind; w_up = w if w < 0, w_up = 0 otherwise",       &
         "1/s",                                                               &
         sigma_switch,                                                        &
         statVarID(115))

    call variable_init("w_up_freq",                                           &
         "frequency of w < 0",                                                &
         "dimensionless (fraction)",                                          &
         sigma_switch,                                                        &
         statVarID(116))


    call check(nf90_enddef(ncfileID))
    
  end subroutine initialize_variables
  
! #############################################################################

  subroutine variable_init(name, long_name, units, switch, VarID)
  
    character(len=*), intent(in) :: name, long_name, units
    integer, intent(in) :: switch
    integer, intent(out) :: VarID

       if(switch == theta_switch) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/ latDimID, thetaDimID /), VarID ) )
       elseif(switch == sigma_switch) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/ latDimID, sigmaDimID /), VarID ) )
       elseif(switch == surface_switch) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/ latDimID /), VarID ) )
       elseif(switch == zon_wave_switch) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/latDimID, zon_wavenDimID /), VarID ) )
       elseif(switch == spectra_switch) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/zon_wavenDimID, legendreDimID /), VarID ) )
       elseif(switch == spectra_switch_3d) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/zon_wavenDimID, legendreDimID, sigmaDimID /), VarID ) )
       elseif(switch == pdf_switch) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/latDimID, sigmaDimID, binDimID /), VarID ) )
       endif
       
       call check(nf90_put_att(ncFileID, VarID,                               &
            "long_name", trim(long_name)))
       call check(nf90_put_att(ncFileID, VarID,                               &
            "units", trim(units)))

     end subroutine variable_init

! #############################################################################

  subroutine write_variables(num_lat,                     num_fourier,        &
                             num_lev,                       num_times )

!----------------------------- input arguments --------------------------------

    integer, intent(in) ::                                                    &
         num_lat,          &   !
         num_fourier,      &   !
         num_lev               !
    
    real, intent(in) ::                                                       &
         num_times

!----------------------------- local variables --------------------------------

    real    :: zon_waven(num_fourier+1), legendre(num_fourier+1)
    integer :: i, j, k

!----------------------------- executable code --------------------------------

!------------------          write coordinate axes         --------------------

    ! latitudes
    call check(nf90_put_var(ncFileID, latVarID, deg_lat,                      &
         start = (/ 1 /), count = (/ num_lat /) ) )

    ! sigma level
    call check(nf90_put_var(ncFileID, sigmaVarID, sigma ) )

    ! potential temperature
    call check(nf90_put_var(ncFileID, thetaVarID, theta_coord ) )

    call check(nf90_put_var(ncFileID, binVarID, rhum_bin) )

    ! zonal wave number
    do i=0, num_fourier
       zon_waven(i+1) = float(i)
       legendre(i+1) = float(i)
    enddo

    call check(nf90_put_var(ncFileID, zon_wavenVarID, zon_waven ) )
    call check(nf90_put_var(ncFileID, legendreVarID, legendre ) )

    ! times
    call check(nf90_put_var(ncFileID, timeVarID, num_times ) )

    ! --------------             write statistics             -----------------

    call check(nf90_put_var(ncFileID, statVarID( 1), sfctn_isavg,             &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID( 2), sfc_frequency,           &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID( 3), d_dy_pot_temp_avg,       &
         start = (/ 1, 1 /), count = (/num_lat, num_lev/) )) 

    call check(nf90_put_var(ncFileID, statVarID( 4), d_dp_pot_temp_avg,       &
         start = (/ 1, 1 /), count = (/num_lat, num_lev/) )) 

    call check(nf90_put_var(ncFileID, statVarID(5), mrdnl_eddy_heatflux,      &
         start = (/ 1, 1 /), count = (/num_lat, num_lev/) )) 

    call check(nf90_put_var(ncFileID, statVarID(6), u_avg,                    &
         start = (/ 1, 1 /), count = (/num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(7), temp_avg,                 &
         start = (/ 1, 1 /), count = (/num_lat, num_lev/) ))
         
    call check(nf90_put_var(ncFileID, statVarID(8), pot_temp_avg,             &
         start = (/ 1, 1 /), count = (/num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(9), buoyancy_avg,             &
         start = (/ 1, 1 /), count = (/num_lat, num_lev/) ))
         
    call check(nf90_put_var(ncFileID, statVarID(10), ucos_zon_spec_avg,       &
         start = (/ 1, 1 /), count = (/ num_lat, num_fourier /) ))

    call check(nf90_put_var(ncFileID, statVarID(11), vcos_zon_spec_avg,       &
         start = (/ 1, 1 /), count = (/ num_lat, num_fourier/) ))

    call check(nf90_put_var(ncFileID, statVarID(12), p_isavg,                 &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(13), p_full_avg,              &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(14), temp_var_avg,            &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(15), sfctn_avg,               &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(16), ipv_flux_isavg,          &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))
         
    call check(nf90_put_var(ncFileID, statVarID(17), dens_isavg,              &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))
         
    call check(nf90_put_var(ncFileID, statVarID(18), ipv_isavg,               &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))
         
    call check(nf90_put_var(ncFileID, statVarID(19), v_isavg,                 &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))
         
    call check(nf90_put_var(ncFileID, statVarID(20), ipv2_isavg,              &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))
         
    call check(nf90_put_var(ncFileID, statVarID(21), pot_temp_var_avg,        &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(22), v_avg,                   &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(23), v_var_avg,               &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(24), u_var_avg,               &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(25), uvcos_eddy,              &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(26), v2_isavg,                &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(27), spv_flux_y_avg,          &
         start = (/ 1 /), count = (/ num_lat /) ))

    call check(nf90_put_var(ncFileID, statVarID(28), exner_isavg,             &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(29), freq_stat_unstable,      &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(30), mrdnl_tempflux_avg,      &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(31), svol_avg,                &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(32), geos_heat_flux_sfc,      &
         start = (/ 1 /), count = (/ num_lat /) ))

    call check(nf90_put_var(ncFileID, statVarID(33), TEM_res_circ,            &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(34), MTEM_res_circ,           &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(35), vrtcl_heatflux_avg,      &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(36), vrtcl_eddy_heatflux,     &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(37), mrdnl_eddy_tempflux,     &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(38), vrtcl_eddy_tempflux,     &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(39),eddy_temp_div,            &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(40), ps_grid_avg,             &
         start = (/ 1 /), count = (/ num_lat /) ))

    call check(nf90_put_var(ncFileID, statVarID(41), w_avg,                   &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(42), vrtcl_tempflux_avg,      &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(43), wu_eddy,                 &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(44), pv_eddy_isavg,           &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(45), pv_theta_eddy_isavg,     &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(46), rhov_eddy_isavg,         &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(47), p_var_isavg,             &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(48), u2_isavg,                &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(49), u_isavg,                 &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(50), u_baro_var_avg,          &
         start = (/ 1 /), count = (/ num_lat /) ))

    call check(nf90_put_var(ncFileID, statVarID(51), v_baro_var_avg,          &
         start = (/ 1 /), count = (/ num_lat /) ))

    call check(nf90_put_var(ncFileID, statVarID(52), pexner_isavg,            &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(53), mont_x_isavg,            &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(54), rho_mont_x_isavg,        &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(55), bc_sfctn_spec,           &
         start = (/ 1, 1 /), count = (/ num_fourier+1, num_fourier+1 /) ))

    call check(nf90_put_var(ncFileID, statVarID(56), bt_sfctn_spec,           &
         start = (/ 1, 1 /), count = (/ num_fourier+1, num_fourier+1 /) ))

    call check(nf90_put_var(ncFileID, statVarID(57), theta_s_mont_x,          &
         start = (/ 1 /), count = (/ num_lat /) ))

    call check(nf90_put_var(ncFileID, statVarID(58), bt_sfctn_avg,            &
         start = (/ 1 /), count = (/ num_lat /) ))

    call check(nf90_put_var(ncFileID, statVarID(59), bt_sfctn_var_avg,        &
         start = (/ 1 /), count = (/ num_lat /) ))

    call check(nf90_put_var(ncFileID, statVarID(60), mont_isavg,              &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(61), mont_var_isavg,          &
         start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(62), pot_temp_spec,           &
         start = (/ 1, 1, 1 /), count = (/ num_fourier+1, num_fourier+1, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(63), conv_spec,               &
         start = (/ 1, 1, 1 /), count = (/ num_fourier+1, num_fourier+1, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(64), hor_sfctn_avg,          &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(65), hor_sfctn_var_avg,      &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    if(moisture) then

       call check(nf90_put_var(ncFileID, statVarID(66), shum_avg,                            &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))
       
       call check(nf90_put_var(ncFileID, statVarID(67), rhum_avg,                            &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))
       
       call check(nf90_put_var(ncFileID, statVarID(68), shum_isavg,                          &
            start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

       call check(nf90_put_var(ncFileID, statVarID(69), sat_shum_isavg,                      &
            start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

       call check(nf90_put_var(ncFileID, statVarID(70), vshum_avg,                           &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))
       
       call check(nf90_put_var(ncFileID, statVarID(71), vshum_isavg,                         &
            start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

       call check(nf90_put_var(ncFileID, statVarID(72), shum_zon_spec_avg,                   &
            start = (/ 1, 1 /), count = (/ num_lat, num_fourier /) ))

       call check(nf90_put_var(ncFileID, statVarID(73), shum_spec,                           &
            start = (/ 1, 1, 1 /), count = (/ num_fourier+1, num_fourier+1, num_lev /) ))

       call check(nf90_put_var(ncFileID, statVarID(74), mrdnl_eddy_theta_e_flux,             &
            start = (/ 1, 1 /), count = (/num_lat, num_lev/) )) 
   
       call check(nf90_put_var(ncFileID, statVarID(75), theta_e_avg,                         &
            start = (/ 1, 1 /), count = (/num_lat, num_lev/) ))
   
       call check(nf90_put_var(ncFileID, statVarID(76), theta_e_var_avg,                     &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))
   
       call check(nf90_put_var(ncFileID, statVarID(77), vrtcl_theta_e_flux_avg,              &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))
   
       call check(nf90_put_var(ncFileID, statVarID(78), vrtcl_eddy_theta_e_flux,             &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))
   
       call check(nf90_put_var(ncFileID, statVarID(79), theta_e_spec,                        &
            start = (/ 1, 1, 1 /), count = (/ num_fourier+1, num_fourier+1, num_lev /) ))
   
       call check(nf90_put_var(ncFileID, statVarID(80), theta_e_sat_avg,                     &
            start = (/ 1, 1 /), count = (/num_lat, num_lev/) ))

       call check(nf90_put_var(ncFileID, statVarID(81), d_dy_theta_e_avg,                    &
            start = (/ 1, 1 /), count = (/num_lat, num_lev/) )) 

       call check(nf90_put_var(ncFileID, statVarID(82), d_dp_theta_e_avg,                    &
            start = (/ 1, 1 /), count = (/num_lat, num_lev/) )) 

       call check(nf90_put_var(ncFileID, statVarID(83), shum_var_avg,                        &
            start = (/ 1, 1 /), count = (/num_lat, num_lev/) )) 

       call check(nf90_put_var(ncFileID, statVarID(84), sat_shum_avg,                        &
            start = (/ 1, 1 /), count = (/num_lat, num_lev/) )) 

       call check(nf90_put_var(ncFileID, statVarID(85), rhum_pdf,                            &
            start = (/ 1, 1 /), count = (/num_lat, num_lev, num_bin/) )) 

       call check(nf90_put_var(ncFileID, statVarID(86), wshum_avg,                           &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

       call check(nf90_put_var(ncFileID, statVarID(87), temp_shum_avg,                       &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

       call check(nf90_put_var(ncFileID, statVarID(88), sat_shum_var_avg,                    &
            start = (/ 1, 1 /), count = (/num_lat, num_lev/) )) 

       call check(nf90_put_var(ncFileID, statVarID(89), vsat_shum_isavg,                     &
            start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))

       call check(nf90_put_var(ncFileID, statVarID(90), rhum_isavg,                          &
            start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))
   
    endif

    call check(nf90_put_var(ncFileID, statVarID(91), z_avg,                               &
         start = (/ 1, 1 /), count = (/num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(92), mrdnl_z_flux_avg,                    &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(93), mrdnl_eddy_z_flux,                   &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    if(ozone) then

       call check(nf90_put_var(ncFileID, statVarID(94), ozone_avg,            &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))
                                                                                                                                     
       call check(nf90_put_var(ncFileID, statVarID(95), ozone_isavg,          &
            start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))
                                                                                                                                     
       call check(nf90_put_var(ncFileID, statVarID(96), vozone_avg,           &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))
                                                                                                                                     
       call check(nf90_put_var(ncFileID, statVarID(97), vozone_isavg,         &
            start = (/ 1, 1 /), count = (/ num_lat, MaxIsentrLev /) ))
                                                                                                                                     
       call check(nf90_put_var(ncFileID, statVarID(98), ozone_zon_spec_avg,   &
            start = (/ 1, 1 /), count = (/ num_lat, num_fourier /) ))
                                                                                                                                     
       call check(nf90_put_var(ncFileID, statVarID(99), ozone_spec,           &
            start = (/ 1, 1, 1 /), count = (/ num_fourier+1, num_fourier+1, num_lev /) ))
                                                                                                                                     
    endif

    if (moisture) then

       call check(nf90_put_var(ncFileID, statVarID(100), sat_shum_spec,       &
            start = (/ 1, 1, 1 /), count = (/ num_fourier+1, num_fourier+1, num_lev /) ))

       call check(nf90_put_var(ncFileID, statVarID(101), rhum_spec,       &
            start = (/ 1, 1, 1 /), count = (/ num_fourier+1, num_fourier+1, num_lev /) ))

    endif

    if(moist_tend) then

       call check(nf90_put_var(ncFileID, statVarID(102), dt_qg_cond_avg,     &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

       call check(nf90_put_var(ncFileID, statVarID(103), dt_qg_conv_avg,     &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

       call check(nf90_put_var(ncFileID, statVarID(104), dt_qg_cond_freq,    &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

       call check(nf90_put_var(ncFileID, statVarID(105), dt_qg_conv_freq,    &
            start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    endif

    call check(nf90_put_var(ncFileID, statVarID(106), w_var_avg,             &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    if(precip) then
       call check(nf90_put_var(ncFileID, statVarID(107), precip_cond_avg,         &
            start = (/ 1 /), count = (/ num_lat /) ))

       call check(nf90_put_var(ncFileID, statVarID(108), precip_conv_avg,         &
            start = (/ 1 /), count = (/ num_lat /) ))

       call check(nf90_put_var(ncFileID, statVarID(109), precip_intensity,        &
            start = (/ 1 /), count = (/ num_lat /) ))

       call check(nf90_put_var(ncFileID, statVarID(110), precip_cond_intensity,   &
            start = (/ 1 /), count = (/ num_lat /) ))

       call check(nf90_put_var(ncFileID, statVarID(111), precip_conv_intensity,   &
            start = (/ 1 /), count = (/ num_lat /) ))
    endif

    call check(nf90_put_var(ncFileID, statVarID(112), vrtcl_z_flux_avg,          &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(113), vrtcl_eddy_z_flux,         &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(114), w_cube_avg,             &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(115), w_up_avg,               &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(116), w_up_freq,              &
         start = (/ 1, 1 /), count = (/ num_lat, num_lev /) ))

    call check(nf90_sync(ncFileID))

  end subroutine write_variables

! #############################################################################

  subroutine close_output_file

    call check(nf90_close(ncFileID))

  end subroutine close_output_file

! #############################################################################

  subroutine check(status)
    
    ! checks error status after each netcdf, prints out text message each time
    !   an error code is returned. 
    
    integer, intent(in) :: status
    
    if(status /= nf90_noerr) then 
       write(*, *) trim(nf90_strerror(status))
    end if
  end subroutine check
  
end module output_mod



