program offlinediag

  use netcdf
  use input_mod, only:     read_data,                open_input_files,        &
                      get_dimensions,                       axes_init,        &
              read_surf_geopotential

  use get_grid_fields_mod, only:                                              &
                    grid_fields_init,                 get_grid_fields,        &
                       surf_gradient,             zonal_spectrum_grid

  use vars

  use isentropic_mod, only:                                                   &
                     isentropic_init,             sigma_to_isentropes,        &
                isentropic_variables,                      isentrdens

  use local_utilities_mod, only:                                              &
                           sigma_avg,        interpolate_half_to_full,        &
                           mrdnl_avg,            update_sfc_frequency,        &
               residual_circulations,              buoyancy_frequency,        &
                      divide_by_psfc,            update_stat_unstable,        &
                      pot_temp_deriv,                  vrtcl_gradient,        &
                      mrdnl_gradient,                       vrtcl_int,        &
                             pdf_avg,                    sigma_avg_3d,        &
                   divide_by_psfc_3d

  use constants_and_switches_mod, only:                                       &
            reference_sea_level_pres,                              cp,        &
                               rdgas,                           kappa,        &
                                  pi,                           omega,        &
                                grav,                          radius,        &
                               rvgas

  use output_mod, only:                                                       &
                  create_output_file,                 write_variables,        &
                   close_output_file


  implicit none

! ------------------------------ local variables ------------------------------

  real :: time                 ! current time
  integer :: time_counter=1    ! used to calculate frequencies based on daily data

  ! local dimensions of gridpoint arrays
  integer ::                                                                  &
       num_lat,            &   ! total number of latitudes
       num_lon,            &   ! total number of longitudes
       num_lev,            &   ! total number of levels
       num_times,          &   ! total number of times
       curr_num_times          ! number of times in current segment   

  integer i,j,k

! ------------------------------ executable code -----------------------------

  call read_namelist

  call open_input_files(      DataIn,                     Data_Source,        &
                            moisture,                           ozone,        &
                          moist_tend,                          precip,        &
                             virtual,                                         &
                            UVarName,                        VVarName,        &
                          VORVarName,                      DIVVarName,        &
                         TEMPVarName,                       PSVarName,        &
                         SHumVarName,                     CondVarName,        &
                         ConvVarName,                 CondTempVarName,        &
                     ConvTempVarName,                  RadTempVarName,        &
                     DiffTempVarName,                 DiffShumVarName,        &
                   PrecipCondVarName,                                         &
                   PrecipConvVarName,                  UInputFileName,        &
                      VInputFileName,                VORInputFileName,        &
                    DIVInputFileName,               TEMPInputFileName,        &
                     PSInputFileName,               SHumInputFileName,        &
                   CondInputFileName,               ConvInputFileName,        &
               CondTempInputFileName,           ConvTempInputFileName,        &
                RadTempInputFileName,           DiffTempInputFileName,        &
               DiffShumInputFileName,                                         &
             PrecipCondInputFileName,         PrecipConvInputFileName)



  call get_dimensions(       num_lat,                         num_lon,        &
                             num_lev,                       num_times )
!  write(*,*) 'Analyzing ',num_times, ' times'

  ! initialize fields based on dimensions
  call allocate_and_initialize(                                               &
                             num_lat,                         num_lon,        &
                             num_lev )

! get axes
  call axes_init(            deg_lat,                         deg_lon,        &
                             sin_lat,                                         &
                          cos_lat_xy,                      cos_lat_yz )

! read_surf_geopotential must be called before grid_fields_init
  if(.not.smooth_surface)  call read_surf_geopotential(                       &
                           SGVarName,                 SGInputFileName,        &
                   surf_geopotential )

! initialize get_grid_fields module
  call grid_fields_init(      DataIn,                     is_gaussian,        &
                            moisture,                           ozone,        &
                    moist_isentropes,                     data_source,        &
                         num_fourier,                         num_lat,        &
                             num_lon,                         num_lev,        &
                          cos_lat_xy,               surf_geopotential,        &
                               sigma,                              pk,        &
                                  bk )


! initialize isentropic module
  call isentropic_init(   PotTempMin,                      PotTempMax,        &
                         theta_coord,                    dtheta_coord )


  curr_num_times = num_times / num_segments
  write(*,*) "Starting time ", TimeIn
  write(*,*) 'Analyzing ', curr_num_times, ' instants.'

!  time = 0.0
  time = TimeIn

!  do while (time < num_times)
  do while (time < TimeIn + curr_num_times)

     time = time + 1.0

     write(6,*) time

     call read_data(            time,                          u_grid,        &
                              v_grid,                        vor_grid,        &
                            div_grid,                       temp_grid,        &
                             ps_grid,                       shum_grid,        &
                                                      dt_qg_cond_grid,        &
                     dt_qg_conv_grid,                 dt_tg_cond_grid,        &
                     dt_tg_conv_grid,                    tdt_rad_grid,        &
                     dt_tg_diff_grid,                 dt_qg_diff_grid,        &
                    precip_cond_grid,                precip_conv_grid)

     ! calculate virtual temperature consistent with the approximation 
     ! used in the fms code; for use in z, mont_x_grid, N^2, svol
     if (virtual .and. moisture) then
      virtual_temp_grid = temp_grid*(1.0+shum_grid*(rvgas/rdgas-1.0))
     else
      virtual_temp_grid = temp_grid
     endif

     call get_grid_fields(    u_grid,                          v_grid,        &
                            vor_grid,                        div_grid,        &
                           temp_grid,                       shum_grid,        &
                          ozone_grid,                                         &
                             ps_grid,                       vcos_grid,        &
                              w_grid,                      omega_grid,        &
                            pot_temp,                                         &
                             theta_e,                     theta_e_sat,        &
                              p_half,                          p_full,        &
                           mont_grid,                     mont_x_grid,        &
                      hor_sfctn_grid,                      sfctn_grid,        &
                       bt_sfctn_grid,                           sigma,        &
                       bc_sfctn_spec,                   bt_sfctn_spec,        &
                       pot_temp_spec,                       conv_spec,        &
                           shum_spec,                   sat_shum_spec,        &
                           rhum_spec,                    theta_e_spec,        &
                          ozone_spec,               virtual_temp_grid,        &
                         geopot_grid,                     sat_mr_grid,        &
                       sat_shum_grid,                       rhum_grid,        &
                                                         sfctn_u_grid)

     ! compute w_up
     w_up_grid      = 0.0
     w_up_freq_grid = 0.0
     where( w_grid < 0.0 )
        w_up_grid = w_grid
        w_up_freq_grid = 1.0
     endwhere

     call update_stat_unstable(sigma,                        pot_temp,        &
                  freq_stat_unstable )

! %%%%%%%%%%%%%%%%%%%%% isentropic %%%%%%%%%%%%%%%%%%%%%%%%%%%
   if(isentropic) then
     
     call isentropic_variables(                                               &
                             sin_lat,                        pot_temp,        &
                              p_full,           p_half(:,:,num_lev+1),        &
                            vor_grid,                 &
                               k_sfc,                          k_asfc,        &
                            p_isentr,                    abs_vor_grid)                                         
     ! find isentropic density implied by linear interpolation 
     ! routine sigma_to_isentropes
     call isentrdens(p_isentr, dens_isentr)
 

    call sigma_to_isentropes(                                             &
                        pot_temp,                          p_half,        &
                        p_isentr,                           k_sfc,        &
                        u_grid,                          u_isentr)

    call sigma_to_isentropes(                                             &
                        pot_temp,                          p_half,        &
                        p_isentr,                           k_sfc,        &
                        v_grid,                          v_isentr) 

    call sigma_to_isentropes(                                             &
                        pot_temp,                          p_half,        &
                        p_isentr,                           k_sfc,        &
                        abs_vor_grid,              abs_vor_isentr)

    call sigma_to_isentropes(                                             &
                        pot_temp,                          p_half,        &
                        p_isentr,                           k_sfc,        &
                 dt_tg_cond_grid,                  dt_cond_isentr)

    call sigma_to_isentropes(                                             &
                        pot_temp,                          p_half,        &
                        p_isentr,                           k_sfc,        &
                 dt_tg_conv_grid,                  dt_conv_isentr)

    call sigma_to_isentropes(                                             &
                        pot_temp,                          p_half,        &
                        p_isentr,                           k_sfc,        &
                    tdt_rad_grid,                   dt_rad_isentr)


   ! calculate isentropic potential vorticity
     where (dens_isentr .gt. 1e-6)
        ipv  = abs_vor_isentr * 1.0 / dens_isentr
     elsewhere
        ipv = 0.0
     endwhere

    call theta_averages

   endif
!%%%%%%%%%%%%%%%%%%%%%%% end isentropic %%%%%%%%%%%%%%%%%%%%%%%%%%%%

     ! average (around latitude circle) sigma-coordinate fields (inc. surface)
     call sigma_averages


  enddo ! end loop over times

!  call average_fields(time)
  call average_fields(time-TimeIn)

 !  call eddy_fields

 ! get Brunt-Vaeisaelae frequency  
  buoyancy_avg = buoyancy_frequency(                                          &
                          p_full_avg,                virtual_temp_avg,        &
                        pot_temp_avg,               d_dp_pot_temp_avg )


  call create_output_file(   num_lat,                         num_lev,        &
                             num_lon                                  )

!  call write_variables(      num_lat,                     num_fourier,        &
!                             num_lev,                            time,        &
!                             num_lon                                        )
  call write_variables(      num_lat,                     num_fourier,        &
                             num_lev,                            time-TimeIn, &
                             num_lon                                        )

  call close_output_file()
  
contains
  
! ----------------------------------------------------------------------------

  subroutine sigma_averages

!                            --- local variables ---

    real, dimension(size(u_grid,1),size(u_grid,2)) :: u_baro, v_baro
    real, dimension(size(u_grid,1),size(u_grid,2),size(u_grid,3)) ::         &
         ucos_grid             ! u * cosine(latitude)

!                            --- executable code ---

!  3D fields:
    ps_mean          = ps_mean           + ps_grid 
    p_mean           = p_mean            + p_full
    z_mean           = z_mean            + sigma_avg_3d(geopot_grid / grav, ps_grid)

    u_mean           = u_mean            + sigma_avg_3d(u_grid, ps_grid) 
    v_mean           = v_mean            + sigma_avg_3d(v_grid, ps_grid)
    w_mean           = w_mean            + sigma_avg_3d(w_grid, ps_grid)
    w_up_mean        = w_up_mean         + sigma_avg_3d(w_up_grid, ps_grid)
    w_up_freq_mean   = w_up_freq_mean    + w_up_freq_grid
    omega_mean       = omega_mean        + sigma_avg_3d(omega_grid, ps_grid)
    temp_mean        = temp_mean         + sigma_avg_3d(temp_grid, ps_grid) 
    virtual_temp_mean = virtual_temp_mean +                                  &
         sigma_avg_3d(virtual_temp_grid, ps_grid) 
    pot_temp_mean    = pot_temp_mean     + sigma_avg_3d(pot_temp, ps_grid)
    
    vu_mean          = vu_mean          + sigma_avg_3d(v_grid*u_grid, ps_grid) 
    wu_mean          = wu_mean          + sigma_avg_3d(w_grid*u_grid, ps_grid)
    utemp_mean       = utemp_mean       + sigma_avg_3d(u_grid*temp_grid, ps_grid)
    vtemp_mean       = vtemp_mean       + sigma_avg_3d(v_grid*temp_grid, ps_grid)
    wtemp_mean       = wtemp_mean       + sigma_avg_3d(w_grid*temp_grid, ps_grid)
    utheta_mean      = utheta_mean      + sigma_avg_3d(u_grid*pot_temp, ps_grid)
    vtheta_mean      = vtheta_mean      + sigma_avg_3d(v_grid*pot_temp, ps_grid)
    wtheta_mean      = wtheta_mean      + sigma_avg_3d(w_grid*pot_temp, ps_grid)
    uz_mean          = uz_mean          + sigma_avg_3d(u_grid*geopot_grid/grav, ps_grid)
    vz_mean          = vz_mean          + sigma_avg_3d(v_grid*geopot_grid/grav, ps_grid)
    wz_mean          = wz_mean          + sigma_avg_3d(w_grid*geopot_grid/grav, ps_grid)

    u2_mean          = u2_mean          + sigma_avg_3d(u_grid*u_grid, ps_grid)
    v2_mean          = v2_mean          + sigma_avg_3d(v_grid*v_grid, ps_grid)
    w2_mean          = w2_mean          + sigma_avg_3d(w_grid*w_grid, ps_grid)
    temp2_mean       = temp2_mean       + sigma_avg_3d(temp_grid*temp_grid, ps_grid)
    pot_temp2_mean   = pot_temp2_mean   + sigma_avg_3d(pot_temp*pot_temp, ps_grid)

    sfctn_avg        = sfctn_avg         +                                    &
        interpolate_half_to_full(mrdnl_avg(sfctn_grid))
    sfctn_u_mean     = sfctn_u_mean      +                                    &
        interpolate_half_to_full(sfctn_u_grid) 

    ! 2d upward vertical velocity
    w_up_avg         = w_up_avg          + sigma_avg(w_up_grid, ps_grid)

    
    if(moisture) then
     shum_mean         = shum_mean           + sigma_avg_3d(shum_grid, ps_grid)
     rhum_mean         = rhum_mean           + sigma_avg_3d(rhum_grid, ps_grid)
     sat_shum_mean     = sat_shum_mean       + sigma_avg_3d(sat_shum_grid, ps_grid)
     ushum_mean        = ushum_mean          + sigma_avg_3d(shum_grid*u_grid, ps_grid)
     vshum_mean        = vshum_mean          + sigma_avg_3d(shum_grid*v_grid, ps_grid)
     wshum_mean        = wshum_mean          + sigma_avg_3d(shum_grid*w_grid, ps_grid)
     shum2_mean        = shum2_mean          + sigma_avg_3d(shum_grid*shum_grid, ps_grid)
    endif

    if (moist_tend) then
       dt_qg_cond_avg   = dt_qg_cond_avg + sigma_avg(dt_qg_cond_grid, ps_grid)
       dt_qg_conv_avg   = dt_qg_conv_avg + sigma_avg(dt_qg_conv_grid, ps_grid)

       dt_qg_diff_mean = dt_qg_diff_mean + sigma_avg_3d(dt_qg_diff_grid, ps_grid)
       dt_qg_conv_mean = dt_qg_conv_mean + sigma_avg_3d(dt_qg_conv_grid, ps_grid)
       dt_qg_cond_mean = dt_qg_cond_mean + sigma_avg_3d(dt_qg_cond_grid, ps_grid)
             
       dt_qg_cond_counter = 0
       where(dt_qg_cond_grid.ne.0)
        dt_qg_cond_counter = 1
       endwhere

       dt_qg_conv_counter = 0
       where(dt_qg_conv_grid.ne.0)
        dt_qg_conv_counter = 1
       endwhere

       dt_qg_cond_freq   = dt_qg_cond_freq + sigma_avg(dt_qg_cond_counter, ps_grid)
       dt_qg_conv_freq   = dt_qg_conv_freq + sigma_avg(dt_qg_conv_counter, ps_grid)

    endif

  
    if (energy_tend) then

       dt_tg_cond_mean = dt_tg_cond_mean + sigma_avg_3d(dt_tg_cond_grid, ps_grid)
       dt_tg_conv_mean = dt_tg_conv_mean + sigma_avg_3d(dt_tg_conv_grid, ps_grid)
       tdt_rad_mean    = tdt_rad_mean    + sigma_avg_3d(tdt_rad_grid, ps_grid)
       dt_tg_diff_mean = dt_tg_diff_mean + sigma_avg_3d(dt_tg_diff_grid, ps_grid)

    endif

    if (eke_tend) then

       Qdtemp_mean     =  Qdtemp_mean     + sigma_avg_3d((dt_tg_cond_grid + dt_tg_conv_grid)*temp_grid, ps_grid)
       Qrtemp_mean     =  Qrtemp_mean     + sigma_avg_3d(tdt_rad_grid*temp_grid, ps_grid)
       Qdifftemp_mean  =  Qdifftemp_mean  + sigma_avg_3d(dt_tg_diff_grid*temp_grid, ps_grid)

       Qdtheta_mean    =  Qdtheta_mean    + sigma_avg_3d((dt_tg_cond_grid + dt_tg_conv_grid)*pot_temp, ps_grid)
       Qrtheta_mean    =  Qrtheta_mean    + sigma_avg_3d(tdt_rad_grid*pot_temp, ps_grid)
       Qdifftheta_mean =  Qdifftheta_mean + sigma_avg_3d(dt_tg_diff_grid*pot_temp, ps_grid)

       uek_mean         = uek_mean         + sigma_avg_3d(u_grid*(u_grid*u_grid + v_grid*v_grid), ps_grid)
       vek_mean         = vek_mean         + sigma_avg_3d(v_grid*(u_grid*u_grid + v_grid*v_grid), ps_grid)

       upot_temp2_mean  = upot_temp2_mean  + sigma_avg_3d(u_grid*pot_temp*pot_temp, ps_grid)
       vpot_temp2_mean  = vpot_temp2_mean  + sigma_avg_3d(v_grid*pot_temp*pot_temp, ps_grid)

       omegalpha_mean   = omegalpha_mean   + sigma_avg_3d(w_grid*temp_grid*rdgas/p_full, ps_grid)

    endif

    if (precip) then

       precip_cond_mean  = precip_cond_mean + precip_cond_grid
       precip_conv_mean  = precip_conv_mean + precip_conv_grid

       ! calculate precipitation intensity based on daily data
       ! (input precipitation data should be 4xday and time averaged rather than instantaneous)

       precip_cond_daily = precip_cond_daily + 0.25*precip_cond_grid
       precip_conv_daily = precip_conv_daily + 0.25*precip_conv_grid
       
       if (time_counter == 4) then  ! reached end of day
       
       
        precip_mask = precip_cond_daily .gt. precip_threshold
        
        precip_cond_intensity = precip_cond_intensity +  precip_cond_daily*precip_mask

        wet_days_cond = wet_days_cond + one_array*precip_mask


        precip_mask = precip_conv_daily .gt. precip_threshold 

        precip_conv_intensity = precip_conv_intensity +  precip_conv_daily*precip_mask

        wet_days_conv = wet_days_conv + one_array*precip_mask

        
        precip_mask = (precip_cond_daily+precip_conv_daily) .gt. precip_threshold
        
        precip_intensity = precip_intensity +                            &
             (precip_cond_daily+precip_conv_daily)*precip_mask
        wet_days = wet_days + one_array*precip_mask

       
        precip_cond_daily = 0
        precip_conv_daily = 0

        time_counter        = 1
       
       else
       
        time_counter = time_counter + 1
       
       endif

    endif

 end subroutine sigma_averages
  

! ----------------------------------------------------------------------------

  subroutine theta_averages

      u_ismean       = u_ismean     + u_isentr 
      v_ismean       = v_ismean     + v_isentr
      
      dens_ismean    = dens_ismean  + dens_isentr
      q_ismean       = q_ismean     + ipv

      qv_ismean      =  qv_ismean   + v_isentr * ipv  * 1.0 / dens_isentr
      qu_ismean      =  qu_ismean   + u_isentr * ipv  * 1.0 / dens_isentr
      q2_ismean      =  q2_ismean   + ipv * ipv  * 1.0 / dens_isentr

! Diabatic terms:
      dt_cond_ismean = dt_cond_ismean + dt_cond_isentr
      dt_conv_ismean = dt_conv_ismean + dt_conv_isentr
      dt_rad_ismean  = dt_rad_ismean  + dt_rad_isentr 

      dt_cond_q_ismean = dt_cond_q_ismean + dt_cond_isentr * ipv  * 1.0 / dens_isentr! no density weghting because of pv
      dt_conv_q_ismean = dt_conv_q_ismean + dt_conv_isentr * ipv * 1.0 / dens_isentr
      dt_rad_q_ismean  = dt_rad_q_ismean  + dt_rad_isentr * ipv  * 1.0 / dens_isentr



  end subroutine theta_averages

! ----------------------------------------------------------------------------

  subroutine average_fields(dof)

!----------------------------- input arguments --------------------------------

    real, intent(in) :: dof ! degrees of freedom

!----------------------------- local variables  -------------------------------
    
    integer j

!----------------------------- executable code  -------------------------------
    
     ! divide averages by number of degrees of freedom in estimates
    ps_grid_avg          = ps_grid_avg / dof
    ps_mean              = ps_mean / dof
    p_mean               = p_mean / dof
    z_mean               = divide_by_psfc_3d(z_mean/dof,           ps_mean)

    ! divide sigma-averaged fields by mean sfc pressure and dof's

    u_mean               = divide_by_psfc_3d(u_mean/dof,           ps_mean)
    v_mean               = divide_by_psfc_3d(v_mean/dof,           ps_mean)
    w_mean               = divide_by_psfc_3d(w_mean/dof,           ps_mean)
    w_up_mean            = divide_by_psfc_3d(w_up_mean/dof,        ps_mean)
    w_up_freq_mean       = w_up_freq_mean / dof
    omega_mean           = divide_by_psfc_3d(omega_mean/dof,       ps_mean)
    temp_mean            = divide_by_psfc_3d(temp_mean/dof,        ps_mean)
    virtual_temp_mean    = divide_by_psfc_3d(virtual_temp_mean/dof, ps_mean)
    pot_temp_mean        = divide_by_psfc_3d(pot_temp_mean/dof,    ps_mean)
    
    vu_mean              = divide_by_psfc_3d(vu_mean/dof,          ps_mean)
    wu_mean              = divide_by_psfc_3d(wu_mean/dof,          ps_mean) 
    utemp_mean           = divide_by_psfc_3d(utemp_mean/dof,       ps_mean)
    vtemp_mean           = divide_by_psfc_3d(vtemp_mean/dof,       ps_mean)
    wtemp_mean           = divide_by_psfc_3d(wtemp_mean/dof,       ps_mean)
    utheta_mean          = divide_by_psfc_3d(utheta_mean/dof,      ps_mean)
    vtheta_mean          = divide_by_psfc_3d(vtheta_mean/dof,      ps_mean)
    wtheta_mean          = divide_by_psfc_3d(wtheta_mean/dof,      ps_mean)
    uz_mean              = divide_by_psfc_3d(uz_mean/dof,          ps_mean)
    vz_mean              = divide_by_psfc_3d(vz_mean/dof,          ps_mean)
    wz_mean              = divide_by_psfc_3d(wz_mean/dof,          ps_mean)

    u2_mean              = divide_by_psfc_3d(u2_mean/dof,          ps_mean)
    v2_mean              = divide_by_psfc_3d(v2_mean/dof,          ps_mean)
    w2_mean              = divide_by_psfc_3d(w2_mean/dof,          ps_mean)
    temp2_mean           = divide_by_psfc_3d(temp2_mean/dof,       ps_mean)
    pot_temp2_mean       = divide_by_psfc_3d(pot_temp2_mean/dof,   ps_mean)

    sfctn_avg            = sfctn_avg / dof

!    don't have ps_avg; one can form the zonal mean w_up in matlab
!    w_up_avg             = divide_by_psfc(w_up_avg/dof, ps_avg)

    if(moisture) then

       shum_mean            = divide_by_psfc_3d(shum_mean/dof,     ps_mean)
       rhum_mean            = divide_by_psfc_3d(rhum_mean/dof,     ps_mean)
       sat_shum_mean        = divide_by_psfc_3d(sat_shum_mean/dof, ps_mean)
       ushum_mean           = divide_by_psfc_3d(ushum_mean/dof,    ps_mean)
       vshum_mean           = divide_by_psfc_3d(vshum_mean/dof,    ps_mean)
       wshum_mean           = divide_by_psfc_3d(wshum_mean/dof,    ps_mean)
       shum2_mean           = divide_by_psfc_3d(shum2_mean/dof,    ps_mean)

    endif

    if(energy_tend) then

    dt_tg_cond_mean        = divide_by_psfc_3d(dt_tg_cond_mean/dof, ps_mean)
    dt_tg_conv_mean        = divide_by_psfc_3d(dt_tg_conv_mean/dof, ps_mean)
    tdt_rad_mean           = divide_by_psfc_3d(tdt_rad_mean/dof,    ps_mean)
    dt_tg_diff_mean        = divide_by_psfc_3d(dt_tg_diff_mean/dof, ps_mean)

    endif

    if (moist_tend) then

       dt_qg_diff_mean        = divide_by_psfc_3d(dt_qg_diff_mean/dof, ps_mean)
       dt_qg_conv_mean        = divide_by_psfc_3d(dt_qg_conv_mean/dof, ps_mean)
       dt_qg_cond_mean        = divide_by_psfc_3d(dt_qg_cond_mean/dof, ps_mean)

    endif

    if(eke_tend) then

    Qdtemp_mean            = divide_by_psfc_3d(Qdtemp_mean/dof,     ps_mean) 
    Qrtemp_mean            = divide_by_psfc_3d(Qrtemp_mean/dof,     ps_mean) 
    Qdifftemp_mean         = divide_by_psfc_3d(Qdifftemp_mean/dof,  ps_mean)
 
    Qdtheta_mean           = divide_by_psfc_3d(Qdtheta_mean/dof,    ps_mean) 
    Qrtheta_mean           = divide_by_psfc_3d(Qrtheta_mean/dof,    ps_mean) 
    Qdifftheta_mean        = divide_by_psfc_3d(Qdifftheta_mean/dof, ps_mean) 

    uek_mean             = divide_by_psfc_3d(uek_mean/dof,          ps_mean)
    vek_mean             = divide_by_psfc_3d(vek_mean/dof,          ps_mean)

    upot_temp2_mean      = divide_by_psfc_3d(upot_temp2_mean/dof,   ps_mean)
    vpot_temp2_mean      = divide_by_psfc_3d(vpot_temp2_mean/dof,   ps_mean)

    omegalpha_mean         = divide_by_psfc_3d(omegalpha_mean/dof,  ps_mean)  

    endif

    if (precip) then
       precip_cond_mean  = precip_cond_mean/dof
       precip_conv_mean  = precip_conv_mean/dof

       precip_intensity      = precip_intensity/wet_days
       precip_cond_intensity = precip_cond_intensity/wet_days_cond
       precip_conv_intensity = precip_conv_intensity/wet_days_conv
    endif

  if (isentropic) then
  dens_ismean  = dens_ismean /dof
  
  u_ismean         =     u_ismean         /dof /dens_ismean
  v_ismean         =     v_ismean         /dof /dens_ismean
  q_ismean         =     q_ismean         /dof /dens_ismean
  qu_ismean        =     qu_ismean        /dof /dens_ismean
  qv_ismean        =     qv_ismean        /dof /dens_ismean
  q2_ismean        =     q2_ismean        /dof /dens_ismean
  dt_cond_ismean   =     dt_cond_ismean   /dof /dens_ismean
  dt_conv_ismean   =     dt_conv_ismean   /dof /dens_ismean
  dt_rad_ismean    =     dt_rad_ismean    /dof /dens_ismean
  dt_cond_q_ismean =     dt_cond_q_ismean /dof /dens_ismean
  dt_conv_q_ismean =     dt_conv_q_ismean /dof /dens_ismean
  dt_rad_q_ismean  =     dt_rad_q_ismean  /dof /dens_ismean

  endif

  end subroutine average_fields


  subroutine read_namelist

    integer :: unit

    unit = 20
    rewind(unit)
    open(unit, file='input.nml')
    read(unit, nml=main_list)
    rewind(unit)
    read(unit, nml=filename_list)

    close(unit)


  end subroutine read_namelist

end program offlinediag
  

