 module output_mod

  use vars
  use typesizes, only: bytesizesOK
  use netcdf
  use constants_and_switches_mod, only: gspval

implicit none
private

public :: create_output_file, write_variables, close_output_file

  integer, parameter ::                                                       &
       nvars = 1002,       &  ! number of output fields (max)
       theta_switch = 1,   &  ! theta variable switch
       sigma_switch = 2,   &  ! sigma variable switch
       surface_switch = 3, &  ! surface variable switch
       zon_wave_switch = 4,&  ! zonal wavenumber variable switch
       spectra_switch = 5, &  ! spectral variable switch
       pdf_switch = 6,     &  ! pdf (lat, lev, bin)
       spectra_switch_3d = 7, &  ! 3d spectral (wn, wn, height)
       sigma_switch_3d = 8,   &  ! sigma variable switch
       surface_switch_3d = 9, &  ! surface variable switch 
       theta_switch_3d = 10      ! surface variable switch   

  integer ::                                                                  &
       ncFileID,           &  ! file ID

       latDimID,           &  ! latitude file ID
       lonDimID,           &  ! longitude file ID
       sigmaDimID,         &  ! sigma dimension ID
       thetaDimID,         &  ! theta dimension ID
       zon_wavenDimID,     &  ! zonal wavenumber dimension ID 
       legendreDimID,      &  ! legendre dimension ID
       binDimID,           &  ! pdf bin dimension ID
       timeDimID,          &  ! time dimension ID
       timeVarID,          &  ! time variable ID
       latVarID,           &  ! latitude variable ID
       lonVarID,           &  ! longitude variable ID
       sigmaVarID,         &  ! sigma variable ID
       thetaVarID,         &  ! theta variable ID
       zon_wavenVarID,     &  ! zonal wavenumber variable ID
       legendreVarID,      &  ! legendre variable ID
       binVarID,           &  ! pdf bin variable ID
       statVarID(nvars)       ! array of field variable IDs

contains
! #############################################################################

  subroutine create_output_file(                                              &
                             num_lat,                         num_lev,        &
                             num_lon                                          )

! creates output file and axis variables

!----------------------------- input arguments --------------------------------
    
    integer, intent(in) :: num_lat, num_lev, num_lon
    
!----------------------------- local variables --------------------------------

!----------------------------- executable code --------------------------------

    if(.not. byteSizesOK()) then
       stop 'open_and_initialize: Compiler does not appear to support required kinds of variables'
    end if

    ! -----------   create netCDF file and define variables  -------------
    ! create netcdf file
    call check(nf90_create(path = trim(OutputFileName),                       &
         cmode = nf90_share, ncid = ncFileID))

    ! global attributes
    call check(nf90_put_att(ncFileID, nf90_global, "title",                   &
         "OfflineDiag Analyses") )
       
       ! global attributes
    call check(nf90_put_att(ncFileID, nf90_global, "title",                   &
         "OfflineDiag Analyses") )


    ! Define axes
    ! -----------
    ! latitude
    call check(NF90_DEF_DIM(ncFileID, "lat", num_lat, latDimID) )
    call check(nf90_def_var(ncFileID, "lat", nf90_double, latDimID,           &
         latVarID) )  
    call check(nf90_put_att(ncFileID, latVarID,                               &
         "long_name", "latitude") )
    call check(nf90_put_att(ncFileID, latVarID,                               &
         "units", "degrees_north") )

    ! longitude
    call check(NF90_DEF_DIM(ncFileID, "lon", num_lon, lonDimID) )
    call check(nf90_def_var(ncFileID, "lon", nf90_double, lonDimID,           &
         lonVarID) )  
    call check(nf90_put_att(ncFileID, lonVarID,                               &
         "long_name", "longitute") )
    call check(nf90_put_att(ncFileID, lonVarID,                               &
         "units", "degrees_east") )
       
    ! sigma level
    call check(NF90_DEF_DIM(ncFileID, "sigma", num_lev, sigmaDimID) )
    call check(nf90_def_var(ncFileID, "sigma", nf90_double, sigmaDimID,       &
         sigmaVarID))
    call check(nf90_put_att(ncFileID, sigmaVarID,                             &
         "long_name", "sigma level") )
    call check(nf90_put_att(ncFileID, sigmaVarID,                             &
         "units", "sigma_level") )
    
    ! potential temperature
    call check(NF90_DEF_DIM(ncFileID, "theta", MaxIsentrLev, thetaDimID) )
    call check(nf90_def_var(ncFileID, "theta", nf90_double, thetaDimID,       &
         thetaVarID))
    call check(nf90_put_att(ncFileID, thetaVarID,                             &
         "long_name", "potential temperature") )
    call check(nf90_put_att(ncFileID, thetaVarID,                             &
         "units", "deg_K") )
    
    ! zonal wavenumber
    call check(nf90_def_dim(ncFileID, "zon_waven", num_fourier+1,             &
         zon_wavenDimID) )
    call check(nf90_def_var(ncFileID, "zon_waven", nf90_double,               &
         zon_wavenDimID, zon_wavenVarID) )
    call check(nf90_put_att(ncFileID, zon_wavenVarID,                         &
         "long_name", "zonal wavenumber") )
    call check(nf90_put_att(ncFileID, zon_wavenVarID,                         &
         "units", "none") )
       
    ! zonal wavenumber
    call check(nf90_def_dim(ncFileID, "legendre", num_fourier+1,              &
         legendreDimID) )
    call check(nf90_def_var(ncFileID, "legendre", nf90_double,                &
         legendreDimID, legendreVarID) )
    call check(nf90_put_att(ncFileID, legendreVarID,                          &
         "long_name", "legendre wavenumber") )
    call check(nf90_put_att(ncFileID, legendreVarID,                          &
         "units", "none") )

    ! pdf bin for relative humidity
    call check(NF90_DEF_DIM(ncFileID, "rhum_bin", num_bin, binDimID) )
    call check(nf90_def_var(ncFileID, "rhum_bin", nf90_double, binDimID,      &
         binVarID))
    call check(nf90_put_att(ncFileID, binVarID,                               &
         "long_name", "relative humidity bin") )
    call check(nf90_put_att(ncFileID, binVarID,                               &
         "units", "none") )
       
    ! time
    call check(NF90_DEF_DIM(ncFileID, "times", 1, timeDimID) )
    call check(nf90_def_var(ncFileID, "times", nf90_double, timeDimID,        &
         timeVarID) )  
    call check(nf90_put_att(ncFileID, timeVarID,                              &
         "long_name", "number of instants") )
       
!    call initialize_variables
    call initialize_variables_3d
   
     end subroutine create_output_file
    
! #############################################################################




  subroutine initialize_variables_3d 
 

     call variable_init("u",                                                    &
            "Zonal wind",                                                       &
            "m/s",                                                              &
            sigma_switch_3d,                                                    &
            statVarID(1)) 

    call variable_init("v",                                                     &
            "Meridional wind",                                                  &
            "m/s",                                                              &
            sigma_switch_3d,                                                    &
            statVarID(2))

    call variable_init("w",                                                     &
            "vertical wind; w = \dot\sigma",                                    &
            "1/s",                                                              &
            sigma_switch_3d,                                                    &
            statVarID(3)) 

    call variable_init("w_up",                                                  &
            "upward vertical wind; w_up = w if w < 0, w_up = 0 otherwise",      &
            "1/s",                                                              &
            sigma_switch_3d,                                                    &
            statVarID(67)) 

    call variable_init("w_up_freq",                                             &
            "frequency of w < 0",                                  &
            "dimensionless (fraction)",                                         &
            sigma_switch_3d,                                                    &
            statVarID(68)) 

!    call variable_init("w_up_zonalmean",                                        &
!            "zonal mean upward vertical wind; w_up = w if w < 0, w_up = 0 otherwise",      &
!            "1/s",                                                              &
!            sigma_switch,                                                    &
!            statVarID(70)) 

    call variable_init("omega",                                                 &
            "omega; omega = dp/dt",                                             &
            "Pa/s",                                                             &
            sigma_switch_3d,                                                    &
            statVarID(62)) 

     call variable_init("temp",                                                 &
            "Temperature",                                                      &
            "K",                                                                &
            sigma_switch_3d,                                                    &
            statVarID(4)) 

     call variable_init("virtual_temp",                                         &
            "Virtual Temperature",                                              &
            "K",                                                                &
            sigma_switch_3d,                                                    &
            statVarID(66)) 

     call variable_init("pot_temp",                                             &
            "Potential Temperature",                                            &
            "K",                                                                &
            sigma_switch_3d,                                                    &
            statVarID(5)) 

     call variable_init("p_full",                                               &
            "Pressure on Sigma Levels",                                         &
            "Pa",                                                               &
            sigma_switch_3d,                                                    &
            statVarID(6))   

     call variable_init("ps",                                                   &
            "Surface pressure",                                                 &
            "Pa",                                                               &
            surface_switch_3d,                                                  &
            statVarID(7))

     call variable_init("streamfctn",                                           &
            "Eulerian mass transport streamfunction",                           &
            "kg/s",                                                             &
            sigma_switch,                                                  &
            statVarID(69))

     call variable_init("vu",                                                   &
            "Merid. mom. flux",                                                 &
            "m^2/s^2",                                                          &
            sigma_switch_3d,                                                    &
            statVarID(8))    

     call variable_init("wu",                                                   &
            "Vert. mom. flux",                                                  &
            "m^2/s^2",                                                          &
            sigma_switch_3d,                                                    &
            statVarID(9))  

     call variable_init("utemp",                                                &
            "Zonal temp flux",                                                  &
            "m/s*K",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(10))        

     call variable_init("vtemp",                                                &
            "Merid. temp flux",                                                 &
            "m/s*K",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(11))    

     call variable_init("wtemp",                                                &
            "Vert. temp flux",                                                  &
            "m/s*K",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(12))  

    call variable_init("utheta",                                                &
            "Zonal heat flux",                                                  &
            "m/s*K",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(13))   

     call variable_init("vtheta",                                               &
            "Merid. heat flux",                                                 &
            "m/s*K",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(14))    

     call variable_init("wtheta",                                               &
            "Vert. heat flux",                                                  &
            "m/s*K",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(15))

     call variable_init("u2",                                                   &
            "Zonal wind squared",                                               &
            "m^2/s^2",                                                          &
            sigma_switch_3d,                                                    &
            statVarID(16)) 

    call variable_init("v2",                                                    &
            "Meridional wind squared",                                          &
            "m^2/s^2",                                                          &
            sigma_switch_3d,                                                    &
            statVarID(17))

    call variable_init("w2",                                                    &
            "vertical wind squared; w = \dot\sigma",                            &
            "1/s^2",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(18))  

    call variable_init("temp2",                                                 &
            "Temperature squared",                                              &
            "K^2",                                                              &
            sigma_switch_3d,                                                    &
            statVarID(19)) 

    call variable_init("streamfctn_u",                                          &
            "Integral of u: zonal Eulerian mass streamfunction",                &
            "kg/s",                                                              &
            sigma_switch_3d,                                                    &
            statVarID(64)) 

    call variable_init("pot_temp2",                                             &
            "Potential Temperature squared",                                    &
            "K^2",                                                              &
            sigma_switch_3d,                                                    &
            statVarID(20)) 

     call variable_init("z_mean",                                               &
            "Geopotential",                                                     &
            "m",                                                                &
            sigma_switch_3d,                                                    &
            statVarID(21))  

    call variable_init("uz",                                                    &
            "correlation of u and geopotential",                                &
            "m^2/s",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(22))    

    call variable_init("vz",                                                    &
            "correlation of v and geopotential",                                &
            "m^2/s",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(23))    

    call variable_init("wz",                                                    &
            "correlation of w and geopotential",                                &
            "m/s",                                                              &
            sigma_switch_3d,                                                    &
            statVarID(24))    

if (moisture) then

     call variable_init("shum",                                                 &
            "Specific humidity",                                                &
            "kg/kg",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(25))  

     call variable_init("sat_shum",                                             &
            "Saturation Specific humidity",                                     &
            "kg/kg",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(61))  

     call variable_init("rhum",                                                 &
            "Relative humidity",                                                &
            "none",                                                             &
            sigma_switch_3d,                                                    &
            statVarID(26))  

    call variable_init("ushum",                                                &
            "Zonal. moisture flux",                                             &
            "kg.m/kg/s",                                                        &
            sigma_switch_3d,                                                    &
            statVarID(27))  

     call variable_init("vshum",                                                &
            "Merid. moisture flux",                                             &
            "kg.m/kg/s",                                                        &
            sigma_switch_3d,                                                    &
            statVarID(28))    

     call variable_init("wshum",                                                &
            "Vert. moisture flux",                                              &
            "kg.m/kg/s",                                                        &
            sigma_switch_3d,                                                    &
            statVarID(29))     
     
     call variable_init("shum2",                                                &
            "Specific humidity squared",                                        &
            "kg^2/kg^2",                                                        &
            sigma_switch_3d,                                                    &
            statVarID(30)) 

     if(moist_tend) then
      call variable_init("dt_q_diff",                                            &
          "specific humidity tendency due to diffusion",                      &
          "1/s",                                                              &
          sigma_switch_3d,                                                    &
          statVarID(63))

      call variable_init("dt_q_conv",                                            &
         "specific humidity tendency due to convection",                      &
         "1/s",                                                               &
         sigma_switch_3d,                                                     &
         statVarID(64))
    
      call variable_init("dt_q_cond",                                            &
         "specific humidity tendency due to condensation",                    &
         "1/s",                                                               &
         sigma_switch_3d,                                                     &
         statVarID(65))   
     endif
endif



if (energy_tend) then

  if (moisture) then
    call variable_init("dt_t_cond",                                             &
            "diabatic temperature tendency due to large scale condensation",     &
            "K/s",                                                              &
            sigma_switch_3d,                                                    &
            statVarID(31))     


    call variable_init("dt_t_conv",                                               &
            "diabatic temperature tendency due to convection",                   &
            "K/s",                                                              &
            sigma_switch_3d,                                                    &
            statVarID(32))    

  endif

    call variable_init("dt_t_rad",                                              &
            "diabatic temperature tendency due to radiation",                    &
            "K/s",                                                              &
            sigma_switch_3d,                                                    &
            statVarID(33))  

    call variable_init("dt_t_diff",                                             &
            "diabatic temperature tendency due to diffusion",                    &
            "K/s",                                                              &
            sigma_switch_3d,                                                    &
            statVarID(34))   
endif

if (isentropic) then

    call variable_init("qi",                                                    &
            "potential vorticity on isentropic surfaces",                       &
            "pv",                                                               &
            theta_switch_3d,                                                    &
            statVarID(35))    

    call variable_init("ui",                                                    &
            "zonal velocity on isentropic surfaces",                            &
            "m/s",                                                              &
            theta_switch_3d,                                                    &
            statVarID(36))  

    call variable_init("vi",                                                    &
            "meridional velocity on isentropic surfaces",                       &
            "m/s",                                                              &
            theta_switch_3d,                                                    &
            statVarID(37))   

    call variable_init("qui",                                                   &
            "zonal pv flux on isentropic surfaces",                             &
            "1/s^2",                                                            &
            theta_switch_3d,                                                    &
            statVarID(38))  

    call variable_init("qvi",                                                   &
            "meridional pv flux on isentropic surfaces",                        &
            "1/s^2",                                                            &
            theta_switch_3d,                                                    &
            statVarID(39))   

    call variable_init("densi",                                                 &
            "density on isentropic surfaces",                                   &
            "Kg/m^3",                                                           &
            theta_switch_3d,                                                    &
            statVarID(40))  

    call variable_init("dt_condi",                                              &
            "temp tendency due to large scale condensation on isentropic surf", &
            "K/s",                                                              &
            theta_switch_3d,                                                    &
            statVarID(41))

    call variable_init("dt_convi",                                              &
            "temp tendency due to convection on isentropic surfaces",           &
            "K/s",                                                              &
            theta_switch_3d,                                                    &
            statVarID(42))

    call variable_init("dt_radi",                                               &
            "temp tendency due to radiation on isentropic surfaces",            &
            "K/s",                                                              &
            theta_switch_3d,                                                    &
            statVarID(43))

    call variable_init("dt_condqi",                                             &
            "dt_condi * pv on isentropic surf",                                 &
            "K/s",                                                              &
            theta_switch_3d,                                                    &
            statVarID(44))

    call variable_init("dt_convqi",                                             &
            "dt_convi * pv on isentropic surfaces",                             &
            "K/s",                                                              &
            theta_switch_3d,                                                    &
            statVarID(45))

    call variable_init("dt_radqi",                                              &
            "dt_radi *pv on isentropic surfaces",                               &
            "K/s",                                                              &
            theta_switch_3d,                                                    &
            statVarID(46))

    call variable_init("q2i",                                                   &
            "pv squared on isentropic surfaces",                                &
            "pv^2",                                                             &
            theta_switch_3d,                                                    &
            statVarID(47))  

endif

if (eke_tend) then

    call variable_init("QdT",                                                   &
            "correlation of diabatic rad and temperature",                      &
            "K^2/s",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(48))    

    call variable_init("QrT",                                                   &
            "correlation of diabatic cond+conv and temperature",                &
            "K^2/s",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(49))    

    call variable_init("QdiffT",                                                 &
            "correlation of diabatic diffusion and temperature",                &
            "K^2/s",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(50))    

    call variable_init("Qdtheta",                                               &
            "correlation of diabatic rad and potential temperature",            &
            "K^2/s",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(51))    

    call variable_init("Qrtheta",                                               &
            "correlation of diabatic cond+conv and potential temperature",      &
            "K^2/s",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(52))   

   call variable_init("Qdifftheta",                                             &
            "correlation of diabatic diffusion and potential temperature",      &
            "K^2/s",                                                            &
            sigma_switch_3d,                                                    &
            statVarID(53))    
 
    call variable_init("uek",                                                   &
            "correlation of u and kinetic energy",                              &
            "m^3/s^2",                                                          &
            sigma_switch_3d,                                                    &
            statVarID(54))    

    call variable_init("vek",                                                   &
            "correlation of v and kinetic energy",                              &
            "m^3/s^2",                                                          &
            sigma_switch_3d,                                                    &
            statVarID(55))    

    call variable_init("upot_temp2",                                            &
            "correlation of u and pot_temp squared",                            &
            "m*K^2/s",                                                          &
            sigma_switch_3d,                                                    &
            statVarID(56))    

    call variable_init("vpot_temp2",                                            &
            "correlation of v and pot_temp squared",                            &
            "m*K^2/s",                                                          &
            sigma_switch_3d,                                                    &
            statVarID(57))    

    call variable_init("omegalpha",                                             &
            "correlation of omega and inverse density",                         &
            "m^2/s^3",                                                          &
            sigma_switch_3d,                                                    &
            statVarID(58))    

endif

if (precip) then 

     call variable_init("precip_cond",                                          &
            "Large scale cond precip intensity",                                &
            "kg/m/m/s",                                                         &
            surface_switch_3d,                                                  &
            statVarID(59))

    call variable_init("precip_conv",                                           &
            "Convection precip intensity",                                      &
            "kg/m/m/s",                                                         &
            surface_switch_3d,                                                  &
            statVarID(60))

endif




    call check(nf90_enddef(ncfileID))

  end subroutine initialize_variables_3d
! #############################################################################

  subroutine variable_init(name, long_name, units, switch, VarID)
  
    character(len=*), intent(in) :: name, long_name, units
    integer, intent(in) :: switch
    integer, intent(out) :: VarID

       if(switch == theta_switch) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/ latDimID, thetaDimID /), VarID ) )
       elseif(switch == sigma_switch) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/ latDimID, sigmaDimID /), VarID ) )
       elseif(switch == surface_switch) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/ latDimID /), VarID ) )
       elseif(switch == sigma_switch_3d) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/ lonDimID, latDimID, sigmaDimID /), VarID ) )
       elseif(switch == surface_switch_3d) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/ lonDimID, latDimID /), VarID ) )
       elseif(switch == theta_switch_3d) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/ lonDimID, latDimID, thetaDimID  /), VarID ) )

! OPEN(91,file='/home/yohai/fms/conv_gray_alb/scripts/my_output.txt')

! WRITE(91,'(14g25.6)'),'DimID', lonDimID, latDimID, sigmaDimID

       elseif(switch == zon_wave_switch) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/latDimID, zon_wavenDimID /), VarID ) )
       elseif(switch == spectra_switch) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/zon_wavenDimID, legendreDimID /), VarID ) )
       elseif(switch == spectra_switch_3d) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/zon_wavenDimID, legendreDimID, sigmaDimID /), VarID ) )
       elseif(switch == pdf_switch) then
          call check(nf90_def_var(ncFileID, trim(name), nf90_double,          &
               (/latDimID, sigmaDimID, binDimID /), VarID ) )
       endif
       
       call check(nf90_put_att(ncFileID, VarID,                               &
            "long_name", trim(long_name)))
       call check(nf90_put_att(ncFileID, VarID,                               &
            "units", trim(units)))

     end subroutine variable_init

! #############################################################################

  subroutine write_variables(num_lat,                     num_fourier,        &
                             num_lev,                       num_times,        &
                             num_lon                                         )

!----------------------------- input arguments --------------------------------

    integer, intent(in) ::                                                    &
         num_lat,          &   !
         num_fourier,      &   !
         num_lev,          &   !
         num_lon           
    
    real, intent(in) ::                                                       &
         num_times

!----------------------------- local variables --------------------------------

    real    :: zon_waven(num_fourier+1), legendre(num_fourier+1)
    integer :: i, j, k

!----------------------------- executable code --------------------------------

!------------------          write coordinate axes         --------------------

    ! latitudes
    call check(nf90_put_var(ncFileID, latVarID, deg_lat,                      &
         start = (/ 1 /), count = (/ num_lat /) ) )

    ! latitudes
    call check(nf90_put_var(ncFileID, lonVarID, deg_lon,                      &
         start = (/ 1 /), count = (/ num_lon /) ) )

    ! sigma level
    call check(nf90_put_var(ncFileID, sigmaVarID, sigma ) )

    ! potential temperature
    call check(nf90_put_var(ncFileID, thetaVarID, theta_coord ) )

    call check(nf90_put_var(ncFileID, binVarID, rhum_bin) )

    ! zonal wave number
    do i=0, num_fourier
       zon_waven(i+1) = float(i)
       legendre(i+1) = float(i)
    enddo

    call check(nf90_put_var(ncFileID, zon_wavenVarID, zon_waven ) )
    call check(nf90_put_var(ncFileID, legendreVarID, legendre ) )

    ! times
    call check(nf90_put_var(ncFileID, timeVarID, num_times ) )

    ! --------------             write statistics             -----------------
! OPEN(91,file='/home/yohai/fms/conv_gray_alb/scripts/my_output.txt')

! WRITE(91,'(14g25.6)'),'u_mean:', size(u_mean, 1), size(u_mean, 2), size(u_mean, 3)
!WRITE(91,'(14g25.6)'),'num:', num_lon, num_lat, num_lev

    call check(nf90_put_var(ncFileID, statVarID(1), u_mean,                    &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(2), v_mean,                    &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(3), w_mean,                    &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 

    call check(nf90_put_var(ncFileID, statVarID(67), w_up_mean,                &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 

    call check(nf90_put_var(ncFileID, statVarID(68), w_up_freq_mean,           &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 

    call check(nf90_put_var(ncFileID, statVarID(62), omega_mean,               &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 

    call check(nf90_put_var(ncFileID, statVarID(4), temp_mean,                 &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(66), virtual_temp_mean,         &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(5), pot_temp_mean,             &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 

    call check(nf90_put_var(ncFileID, statVarID(6), p_mean,                    &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(7), ps_mean,                   &
            start = (/ 1, 1 /), count = (/num_lon,  num_lat /) ))

    call check(nf90_put_var(ncFileID, statVarID(69), sfctn_avg,                &
            start = (/ 1, 1 /), count = (/num_lat,  num_lev /) ))

!    call check(nf90_put_var(ncFileID, statVarID(70), w_up_avg,                 &
!            start = (/ 1, 1 /), count = (/num_lat,  num_lev /) ))

    call check(nf90_put_var(ncFileID, statVarID(8), vu_mean,                   &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(9), wu_mean,                   &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(10),utemp_mean,                &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(11),vtemp_mean,                &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(12), wtemp_mean,               &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 

    call check(nf90_put_var(ncFileID, statVarID(13), utheta_mean,           &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 

    call check(nf90_put_var(ncFileID, statVarID(14), vtheta_mean,              &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(15), wtheta_mean,              &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(16), u2_mean,                  &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(17), v2_mean,                  &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(18), w2_mean,                  &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 

    call check(nf90_put_var(ncFileID, statVarID(19), temp2_mean,               &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(64), sfctn_u_mean,             &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(20), pot_temp2_mean,           &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 

    call check(nf90_put_var(ncFileID, statVarID(21), z_mean,                   &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(22), uz_mean,                  &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 
   
    call check(nf90_put_var(ncFileID, statVarID(23), vz_mean,                  &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(24), wz_mean,                  &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))
   
 
if (moisture) then

    call check(nf90_put_var(ncFileID, statVarID(25), shum_mean,                &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(61), sat_shum_mean,            &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(26), rhum_mean,                &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(27), ushum_mean,               &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(28), vshum_mean,               &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(29), wshum_mean,               &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(30), shum2_mean,               &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(63), dt_qg_diff_mean,          &
         start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))
    call check(nf90_put_var(ncFileID, statVarID(64), dt_qg_conv_mean,          &
         start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))
    call check(nf90_put_var(ncFileID, statVarID(65), dt_qg_cond_mean,          &
         start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))
endif

if (energy_tend) then

  if (moisture) then

    call check(nf90_put_var(ncFileID, statVarID(31), dt_tg_cond_mean,          &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

 
    call check(nf90_put_var(ncFileID, statVarID(32), dt_tg_conv_mean,          &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

  endif

    call check(nf90_put_var(ncFileID, statVarID(33), tdt_rad_mean,             &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

    call check(nf90_put_var(ncFileID, statVarID(34), dt_tg_diff_mean,          &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

endif

if (isentropic) then

    call check(nf90_put_var(ncFileID, statVarID(35), q_ismean,                 &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(36), u_ismean,                 &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(37), v_ismean,                 &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) )) 

    call check(nf90_put_var(ncFileID, statVarID(38), qu_ismean,                &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(39), qv_ismean,                &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(40), dens_ismean,              &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(41), dt_cond_ismean,           &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) )) 

    call check(nf90_put_var(ncFileID, statVarID(42), dt_conv_ismean,           &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(43), dt_rad_ismean,            &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(44), dt_cond_q_ismean,         &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) )) 

    call check(nf90_put_var(ncFileID, statVarID(45), dt_conv_q_ismean,         &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(46), dt_rad_q_ismean,          &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) ))

    call check(nf90_put_var(ncFileID, statVarID(47), q2_ismean,                &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, MaxIsentrLev /) ))
endif

if (eke_tend) then

   call check(nf90_put_var(ncFileID, statVarID(48), Qdtemp_mean,            &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 
   
   call check(nf90_put_var(ncFileID, statVarID(49), Qrtemp_mean,            &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 

   call check(nf90_put_var(ncFileID, statVarID(50), Qdifftemp_mean,         &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 
   
   call check(nf90_put_var(ncFileID, statVarID(51), Qdtheta_mean,           &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 
   
   call check(nf90_put_var(ncFileID, statVarID(52), Qrtheta_mean,           &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

   call check(nf90_put_var(ncFileID, statVarID(53), Qdifftheta_mean,        &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))
   
   call check(nf90_put_var(ncFileID, statVarID(54), uek_mean,               &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 
   
   call check(nf90_put_var(ncFileID, statVarID(55), vek_mean,               &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))
   

   call check(nf90_put_var(ncFileID, statVarID(56), upot_temp2_mean,        &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) )) 
   
   call check(nf90_put_var(ncFileID, statVarID(57), vpot_temp2_mean,        &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

   call check(nf90_put_var(ncFileID, statVarID(58), omegalpha_mean,         &
          start = (/ 1, 1, 1 /), count = (/num_lon, num_lat, num_lev/) ))

endif

if (precip) then

    call check(nf90_put_var(ncFileID, statVarID(59), precip_cond_mean,    &
            start = (/ 1, 1 /), count = (/num_lon,  num_lat /) ))

    call check(nf90_put_var(ncFileID, statVarID(60), precip_conv_mean,    &
            start = (/ 1, 1 /), count = (/num_lon,  num_lat /) ))

endif

  call check(nf90_sync(ncFileID))

  end subroutine write_variables

! #############################################################################

  subroutine close_output_file

    call check(nf90_close(ncFileID))

  end subroutine close_output_file

! #############################################################################

!!$  subroutine read_output_namelist
!!$
!!$    integer :: unit
!!$
!!$    unit = 20
!!$    rewind(unit)
!!$    open(unit, file='output.nml')
!!$    read(unit, nml=output_list)
!!$    rewind(unit)
!!$    read(unit, nml=filename_list)
!!$
!!$    close(unit)
!!$
!!$
!!$  end subroutine read_output_namelist

  subroutine check(status)
    
    ! checks error status after each netcdf, prints out text message each time
    !   an error code is returned. 
    
    integer, intent(in) :: status
    
    if(status /= nf90_noerr) then 
       write(*, *) trim(nf90_strerror(status))
    end if
  end subroutine check
  
end module output_mod



