module vertical_advection_mod

implicit none
private

integer :: SECOND_ORDER=1, FOURTH_ORDER=2, VAN_LEER=3

public :: vertical_advection, SECOND_ORDER, FOURTH_ORDER, VAN_LEER

!=======================================================================
! description of public interfaces
!=======================================================================

! subroutine vertical_advection(dqdt, q, w, zz, dt, scheme)
!
!    adds the tendency due to vertical advection of the quantity (q) to the 
!          the pre-existing tendendy :  dqdt -> dqdt - w*dq/dz
!    designed for use in spectral models, where vertical advection is in 
!         advective, as opposed to flux, form
!
! input:  
!
!  q
!      real, dimension(:,:,:) or dimension(:,:) or dimension(:) 
!      the 3, 2, or 1d field being advected 
!      in the 3 and 2d cases, the vertical dimension is assumed to be the last dimension
!      units: arbitrary = [Q]
!
!  zz
!      real, dimension(:,:,:) or dimension(:,:) or dimension(:) 
!      the position of the vertical boundaries of the grid boxes measured with some
!        vertical coordinate (referred to as 'z')
!      in the 3d case, size(zz,1)= size(q,1), size(zz,2)=size(q,2), size(zz,3) = size(q.3+1)
!       (similarly in the 1 and 2d cases)
!      units: arbitrary = [Z] 
!      can be any vertical coordinate (height, pressure, sigma, etc)
!      zz(:,:,k) is an increasing function of k  (for pressure or sigma coordinates, 
!             k=1 thus refers to the top of the model)
!
!  dt
!      real 
!      the time step 
!      units: arbitrary = [T]
!
!  w
!      real, dimension(:,:,:) or dimension(:,:) or dimension(:) 
!      shape conforms to that of zz
!      the advecting velocity (in 'z'-coordinates) defined at these vertical boundaries
!      w(:,:,1) and w(:,:,size(w,3)) are not assumed to be zero 
!      units: [Z]/[T]
!
! integer :: scheme
!      must be one of: SECOND_ORDER, FOURTH_ORDER, VAN_LEER
!
!   1: second-order centered
!   2: fourth-order centered
!   3: piecewise linear, finite volume (van Leer)
!
! input/output
!
!  dqdt 
!     real, dimension(:,:,:) or dimension(:,:) or dimension(:) 
!     shape conforms to that of q
!     the derivative of q with respect to t
!     the contribution of vertical advection is added to the input
!     units:  [Q]/[T]




!=============================================================

interface vertical_advection
   module procedure vertical_advection_3d
   module procedure vertical_advection_2d
   module procedure vertical_advection_1d
end interface

character(len=128), parameter :: version = '$Id: vertical_advection.f90,v 1.2 2002/02/22 19:06:41 fms Exp $'
character(len=128), parameter :: tagname = '$Name: havana $'
logical :: entry_to_logfile_done = .false.

contains

!===============================================================

subroutine vertical_advection_3d(dqdt, q, w, zz, dt, scheme)

real   , intent(in),    dimension(:,:,:) :: q, w, zz
real   , intent(in)                      :: dt
integer, intent(in)                      :: scheme
real   , intent(inout), dimension(:,:,:) :: dqdt

real, dimension(size(q,1),size(q,2),size(q,3))   :: s, dz
real, dimension(size(q,1),size(q,2),size(q,3)+1) :: q_mid

integer :: n, k, log_unit
real :: c1, c2
character(len=8) :: chtmp


if(scheme /= SECOND_ORDER .and. scheme /= FOURTH_ORDER .and. scheme /= VAN_LEER) then
  write(*,*) 'vertical_advection_3d: scheme must be one of: SECOND_ORDER, FOURTH_ORDER, VAN_LEER, but scheme=', scheme
endif
if( size(w,3).ne.1 + size(q,3) ) then
   write(*,*) 'vertical_advection_3d: shapes of q snd w are not consistent'
endif
if( size(zz,3).ne.1 + size(q,3) ) then
   write(*,*) 'vertical_advection_3d: shapes of q snd zz are not consistent'
endif

n = size(q,3)

dz = zz(:,:,2:n+1) - zz(:,:,1:n)


! the  3 schemes differ only in the value of the field q assigned to the 
!     boundary locations

if(scheme == SECOND_ORDER) then
  do k = 2, n
    q_mid(:,:,k) = 0.5*(q(:,:,k) + q(:,:,k-1))
  end do
endif

if(scheme == FOURTH_ORDER) then
  c1 = 7./12.
  c2 = 1./12.
  do k = 3, n-1
    q_mid(:,:,k) = c1*(q(:,:,k) + q(:,:,k-1)) - c2*(q(:,:,k+1) + q(:,:,k-2))
  end do
  q_mid(:,:,2) = 0.5*(q(:,:,2) + q(:,:,1  ))
  q_mid(:,:,n) = 0.5*(q(:,:,n) + q(:,:,n-1))
endif

if(scheme == VAN_LEER) then
  call slope_box_vert(s, q, dz)

  do k = 2, n
    where(w(:,:,k) >= 0.0)
      q_mid(:,:,k) = (q(:,:,k-1) + 0.5*s(:,:,k-1)*(1.0 - dt*w(:,:,k)/dz(:,:,k-1)))
    elsewhere
      q_mid(:,:,k) = (q(:,:,k)   - 0.5*s(:,:,k  )*(1.0 + dt*w(:,:,k)/dz(:,:,k  )))
    end where
  end do
endif

q_mid(:,:,1  ) = q(:,:,1)  ! if w = 0 at the top and bottom this value is not relevant
q_mid(:,:,n+1) = q(:,:,n)  ! ditto

do k = 1, n
  dqdt(:,:,k) = dqdt(:,:,k)                                           &
     - (  w(:,:,k+1)*(q_mid(:,:,k+1) - q    (:,:,k))                  &
        + w(:,:,k)  *(q    (:,:,k)   - q_mid(:,:,k)) )/dz(:,:,k)      
end do

return
end subroutine vertical_advection_3d

!===============================================================

subroutine vertical_advection_2d(dqdt, q, w, zz, dt, scheme)

real   , intent(in),    dimension(:,:) :: q, w, zz
real   , intent(in)                    :: dt
integer, intent(in)                    :: scheme
real   , intent(inout), dimension(:,:) :: dqdt

real, dimension(size(q   , 1), 1, size(q    , 2)) :: q_3d
real, dimension(size(w   , 1), 1, size(w    , 2)) :: w_3d
real, dimension(size(zz   ,1), 1, size(zz    ,2)) :: zz_3d
real, dimension(size(dqdt ,1), 1, size(dqdt  ,2)) :: dqdt_3d

q_3d     (:,1,:) = q     (:,:)
w_3d     (:,1,:) = q     (:,:)
zz_3d    (:,1,:) = zz    (:,:)
dqdt_3d  (:,1,:) = dqdt  (:,:)

call vertical_advection_3d (dqdt_3d, q_3d, w_3d, zz_3d, dt, scheme)

dqdt(:,:) = dqdt_3d(:,1,:)

return
end subroutine vertical_advection_2d

!===============================================================

subroutine vertical_advection_1d(dqdt, q, w, zz, dt, scheme)

real   , intent(in),    dimension(:) :: q, w, zz
real   , intent(in)                  :: dt
integer, intent(in)                  :: scheme
real   , intent(inout), dimension(:) :: dqdt

real, dimension(1, 1, size(q   )) :: q_3d
real, dimension(1, 1, size(w   )) :: w_3d
real, dimension(1, 1, size(zz  )) :: zz_3d
real, dimension(1, 1, size(dqdt)) :: dqdt_3d

q_3d     (1,1,:) = q    
w_3d     (1,1,:) = q     
zz_3d    (1,1,:) = zz    
dqdt_3d  (1,1,:) = dqdt  

call vertical_advection_3d (dqdt_3d, q_3d, w_3d, zz_3d, dt, scheme)

dqdt = dqdt_3d(1,1,:)

return
end subroutine vertical_advection_1d

!=================================================================

subroutine slope_box_vert(slope, q, dz)

real, intent(in),  dimension(:,:,:) :: q
real, intent(in),  dimension(:,:,:) :: dz
real, intent(out), dimension(:,:,:) :: slope

real, dimension(size(q,1),size(q,2),size(q,3)) :: grad, q_min, q_max
integer :: n, i, j, k

n = size(q,3)

do k = 2, n
  grad(:,:,k) = 2.0*(q(:,:,k)-q(:,:,k-1))/(dz(:,:,k)+dz(:,:,k-1))
end do

do k = 2,n-1
  slope(:,:,k) = (grad(:,:,k+1) + grad(:,:,k))*dz(:,:,k)/2
end do

slope(:,:,1) = grad(:,:,2)*dz(:,:,1)
slope(:,:,n) = grad(:,:,n)*dz(:,:,n)

! monotonicity check 
!   (make sure that slope does not imply values greater than surrounding values)

q_min(:,:,2:n-1) = min(q(:,:,1:n-2),q(:,:,2:n-1),q(:,:,3:n))
q_min(:,:,1)     = min(q(:,:,1)    ,q(:,:,2))
q_min(:,:,n)     = min(q(:,:,n-1)  ,q(:,:,n))

q_max(:,:,2:n-1) = max(q(:,:,1:n-2),q(:,:,2:n-1),q(:,:,3:n))
q_max(:,:,1)     = max(q(:,:,1)    ,q(:,:,2))
q_max(:,:,n)     = max(q(:,:,n-1)  ,q(:,:,n))

slope = sign(1.0,slope) * min( abs(slope), 2.0*(q - q_min), 2.0*(q_max - q))

return
end subroutine slope_box_vert


end module vertical_advection_mod
