module grid_physics

  ! GRID_PHYSICS provides a number of routines that can be combined to
  ! form a "physics" package in an idealized GCM.
  !
  ! The namelist GRID_PHYS.NML contains flags for the kind of  
  ! parameterization to be employed and specific parameters to
  ! be set. The following flags define the kind of parameterization: 
  !        
  !    MOIST_TROPICS:  If FALSE, Newtonian relaxation towards a
  !                    background state as in Held/Suarez (1994)
  !                    provides the diabatic forcing and Rayleigh
  !                    friction near the surface mimicks turbulent
  !                    drag. 
  !
  !                    If TRUE, quadratic friction near the surface
  !                    mimicks turbulent drag. In a narrow region
  !                    around the equator, temperatures are relaxed
  !                    towards a pseudoadiabat that starts at
  !                    the in situ surface temperature and ends at the
  !                    level of neutral buoyancy. For the diabatic
  !                    forcing in the rest of the atmosphere,
  !                    including the "tropical" region, two options
  !                    are available, distinguished from each other by
  !                    the flag GRAY_ATMOSPHERE:  
  !
  !                    Changed slightly by cw 10/7/02: MOIST_TROPICS
  !                    still performs the same role, but now acts in
  !                    main program, determining which of grid_simple_physics
  !                    or hs_forcing gets called.  rayleigh_damping and
  !                    newtonian_damping have been removed from this
  !                    file, since they are present in hs_forcing.
  !                    
  !  GRAY_ATMOSPHERE:  If FALSE, Newtonian relaxation towards the
  !                    radiative equilibrium state of a gray
  !                    atmosphere provides diabatic forcing. The
  !                    radiative equilibrium state is specified so as
  !                    to a have a surface temperature as defined in
  !                    subroutine SURFACE_TEMPERATURE_FORCED and a
  !                    constant skin ("stratosphere") temperature T_STRAT.
  !
  !                    If TRUE, heating rates are obtained by a
  !                    two-stream radiative transfer calculation for a
  !                    gray atmosphere. The surface temperature is
  !                    specified as a lower boundary condition in
  !                    SURFACE_TEMPERATURE_FORCED. The total optical
  !                    thickness of the atmosphere is specified as
  !                    function of latitude in GRAY_OPTICAL_THICKNESS. 
  !
  ! The surface temperature either of the background state for
  ! Newtonian relaxation or in the radiative transfer calculation is
  ! specified by the two parameters DELH and TSFC_MEAN, the
  ! pole-to-equator temperature difference and the mean surface
  ! temperature, respectively. The extent of the frictional boundary
  ! layer is determined by SIGMA_B. For the Newtonian relaxation
  ! parameterizations, SIGMA_B also determines the height of the
  ! low-latitudinal layer in which the relaxation coefficient is
  ! increased compared with the interior atmospheric relaxation
  ! coefficient.  

use     constants_mod, only: kappa, cp_air, grav, pi, rdgas, rvgas, stefan, &
                             hlv, tfreeze

use           fms_mod, only: error_mesg, fatal, file_exist,       &
                             open_namelist_file, check_nml_error, &
                             mpp_pe, mpp_root_pe, close_file,     &
                             write_version_number, stdlog

use  time_manager_mod, only: time_type
use  diag_manager_mod, only: register_diag_field, send_data
!
  implicit none

  
  real, private ::             & ! constants that appear in parametric
       delh          =    60., & ! expression for the radiative equilibrium surface 
       tsfc_mean     =   300., & ! equatorial temperature
       reference_sea_level_press = 101325.

  real, private ::             & ! parameters in Held/Suarez (1994) parameterization:
       delv          =    10., & ! parameter that sets vertical stability     
       t_strat       =   200., & ! "skin" temperature of atmosphere
       ka_days       =    50., & ! latitude-independent part of Newtonian 
       ka,                     & ! relaxation coefficient
       ks_days       =     4., & ! Newtonian damping coefficient in low
       ks,                     & ! latitudes near surface
       kf_days       =     1., & ! Rayleigh friction coefficient in boundary layer
       kf,                     &
       day           = 86400.    ! added by cw 10/8

  real, public ::              &
       sigma_b         = 0.85    ! extent of frictional `PBL'


  logical, private ::            &! switch between gray atmosphere
       gray_atmosphere = .FALSE.,&! radiative transfer and Newtonian relaxation
       fix_tropic_surface_temperature = .false.,&! fix the surface temperature
       tropical_heating = .true.  ! relax to pseudoadiabat in tropics

        
  real, private ::               & ! "mock-moist tropics/gray extratropics" parameters 
       scale_height_ratio = 3.5, & ! ratio of water vapor and pressure scale heights
       max_optical_thickness,    & ! maximum (equatorial) optical thickness of atmosphere
       min_optical_thickness,    & ! minimum (polar) optical thickness of atmosphere
       kt_tropic_days     = 4.,  & ! Newtonian damping time scale in "tropics"
       kt_tropic,                &
       width_tropics_deg  = 6.,  & ! half-width of zone with relaxation towards 
       width_tropics_rad,        & ! moist adiabat
       lapse_scale_tropic = 1.,  & ! scale factor for moist adiabatic lapse rate
       Cdrag           = 7.e-6,  & ! quadratic "drag coefficient" [1/m]
       alpha           = 1.0  ,  & ! factor in surf. temp. calculation
       off_eq          = 0.0  ,  & ! amount to displace heating off equator 
       fix_tropic_temp     = 300.0 ! fixed surface temp in tropics (see param
                                   ! fix_tropic_surface_temperature)
  
  
       namelist /grid_phys_list/   tsfc_mean,        delh,         t_strat, &
              width_tropics_deg,        delv,     ka_days,         ks_days, &
             lapse_scale_tropic,     kf_days,     sigma_b,  kt_tropic_days, &
      reference_sea_level_press,       Cdrag,     gray_atmosphere, &
             scale_height_ratio,   fix_tropic_temp, &
          min_optical_thickness,    max_optical_thickness,                  &
          fix_tropic_surface_temperature,         tropical_heating, alpha,  &
          off_eq


  private grid_phys_list

   character(len=128) :: version='$Id: grid_physics_simple.f90 $'
   character(len=128) :: tag='homemade'
   character(len=14)  :: mod_name = "grid_physics"

! diagnostic IDs
   integer id_tdt, id_tropical_tdt, id_extratropical_tdt, id_teq
   real :: missing_value = -1.e10
contains
  
  !-------------------------------------------------------------------------------

  subroutine grid_phys_init(axes, Time)
    ! reads namelist file with parameters and echoes parameter values

    integer, intent(in) :: axes(4)
    type(time_type), intent(in) :: Time

    integer  unit, io, ierr

      if (file_exist('input.nml')) then
         unit = open_namelist_file ( )
         ierr=1; do while (ierr /= 0)
            read  (unit, nml=grid_phys_list, iostat=io, end=100)
            ierr = check_nml_error (io, 'grid_phys_list')
         enddo
  100     call close_file (unit)
      endif

!     ----- write version info and namelist to log file -----

      call write_version_number (version,tag)
      if (mpp_pe() == mpp_root_pe()) write (stdlog(),nml=grid_phys_list)

!     ---------------write info to screen--------------------

      if(gray_atmosphere) &
           call error_mesg('grid_phys_init',                             &
           'gray_atmosphere is not currently enabled.  Fix calls to surface_temperature_forced',fatal)
      
      
      if(mpp_pe()==mpp_root_pe()) then
         write(*, *) 'Parameters in radiative equilibrium surface temperature:'
         write(*,10) 'Mean surface temperature:         ', tsfc_mean, 'K'
         write(*,10) 'Equator-pole temperature difference:', delh, 'K'
         write(*, *) 'Temperature varies in latitude as:'
         write(*, *) 'tsfc_mean + delh * (1. / 3. - sin^2(lat))'
      endif
      ! moist_tropics now acts in main program removed by cw 10/8
!!$    if (moist_tropics) then
      if (gray_atmosphere) then
         if(mpp_pe()==mpp_root_pe()) then
            write(*, *) 'Non-scattering gray atmosphere in extratropics.'
            write(*,10) 'Maximum optical thickness of atmosphere:        ',  &
                 max_optical_thickness
            write(*,10) 'Minimum optical thickness of atmosphere:        ',  &
                 min_optical_thickness
         endif
      else
         if(mpp_pe()==mpp_root_pe()) then
            write(*, *) 'Relaxation towards background state in extratropics.'
            write(*, *) 'Background state is the radiative equilibrium state'
            write(*, *) 'of a gray atmosphere with an isothermal stratosphere'
            write(*,10) 'at temperature t_strat =', t_strat, 'K.'
            write(*, *)
            write(*,10) 'Relaxation time scale:                              ', &
                 ka_days, 'days'
            write(*,10) 'Relaxation time scale in low latitudes near surface:', &
                 ks_days, 'days'
         endif
         ka = 1./(ka_days * day)
         ks = 1./(ks_days * day)
      endif
      
      if(mpp_pe()==mpp_root_pe()) then
      write(*, *)
      write(*,10) 'Ratio of absorber scale height over pressure scale height:',&
           scale_height_ratio
      write(*, *)
   endif
   if(tropical_heating) then
      if(mpp_pe()==mpp_root_pe()) then
         write(*, *) 'Relaxation towards moist adiabat in tropics.'
         write(*,10) 'Half-width of tropical zone:', width_tropics_deg, 'degrees'
         write(*,10) 'Relaxation time scale:      ', kt_tropic_days, 'days'
         write(*,10) 'Lapse rate scale factor:    ', lapse_scale_tropic
      endif
         width_tropics_rad = width_tropics_deg * pi/180.
         if(fix_tropic_surface_temperature) then
            if(mpp_pe()==mpp_root_pe()) then
               write(*,10) 'Surface temperature in tropics fixed at ',fix_tropic_temp, 'degrees'
            endif
         endif
         kt_tropic = 1. / (kt_tropic_days * day)
      endif
      if(mpp_pe()==mpp_root_pe()) then
         write(*, *)
         write(*, *) 'Quadratic friction in boundary layer.'
         write(*,10) 'Extent of frictional layer: sigma_b =', sigma_b
         write(*,20) 'Drag coefficient:             Cdrag =', Cdrag, 'm**(-1)'
         write(*, *)
      endif
    
10  format(1x,a,1x,f8.2,1x,a)
20  format(1x,a,1x,e8.2,1x,a)

! register fields with diagnostic manager
      id_teq = register_diag_field( mod_name, 'teq',                         &
           axes(1:3), Time, 'equilibrium temperature', 'deg_K',              &
           missing_value=missing_value)

      id_tdt = register_diag_field( mod_name, 'tdt_total_grid_physics',      &
           axes(1:3), Time, 'total temperature tendency', 'deg_K/sec',       &
           missing_value=missing_value)

      id_tropical_tdt = register_diag_field( mod_name,                       &
           'tdt_tropical_grid_physics', axes(1:3), Time,                     &
           'tropical temperature tendency', 'deg_K/sec',                     &
           missing_value=missing_value)

      id_extratropical_tdt = register_diag_field( mod_name,                  &
           'tdt_extratropical_grid_physics', axes(1:3), Time,                &
           'tropical temperature tendency', 'deg_K/sec',                     &
           missing_value=missing_value)

  end subroutine grid_phys_init
  
  !------------------------------------------------------------------

  subroutine compute_grid_physics( is,       ie,       js,       je,      &
           Time, delta_t, lat,    dt_ug,    dt_vg,    dt_tg,      &   
                     dt_trg,       ug,       vg,       tg,      &
                         trg,   p_half,   p_full,                &
                ug_previous,        vg_previous,        tg_previous,      &
                trg_previous,    p_half_previous,      &
            p_full_previous)

    integer, intent(in)                     :: is, ie, js, je
    type(time_type), intent(in)            ::  Time
    real, intent(in)                       ::  delta_t
    real, intent(in), dimension (:,:,:)     :: ug, vg, tg, p_half, p_full
    real, intent(in), dimension (:,:,:,:)   :: trg
    real, intent(in), dimension (:,:)       :: lat
    
    real, optional, dimension (:,:,:)       :: ug_previous, vg_previous, & 
         tg_previous, p_half_previous, p_full_previous
    real, optional, dimension (:,:,:,:)     :: trg_previous
    
    real, intent(inout), dimension(:,:,:)   :: dt_ug, dt_vg, dt_tg
    real, intent(inout), dimension(:,:,:,:) :: dt_trg


    if(present(ug_previous)) then
          call surface_drag(ug_previous, vg_previous, p_half_previous, & 
               p_full_previous, dt_ug, dt_vg)
          call diabatic_forcing(Time, is, js, tg_previous,                 &
               p_half_previous, p_full_previous, lat, dt_tg)
    else
       call error_mesg('compute_grid_physics',                             &
            'ug_previous is not present',fatal)
    end if

  end subroutine compute_grid_physics

  !---------------------------------------------------------------------
   
  subroutine surface_drag(ug, vg, p_half, p_full, dt_ug, dt_vg)
    
    real, intent(in), dimension (:,:,:)    :: ug, vg, p_full, p_half
    real, intent(inout), dimension(:,:,:)  :: dt_ug, dt_vg
    real, dimension(size(ug,1),size(ug,2)) :: sigma, sigma_norm, sigma_max
    real, dimension(size(ug, 1), size(ug, 2), size(ug, 3)) :: vg_norm
    integer :: k, num_level
    
    ! pseudo=parameter for number of levels
    num_level = size(ug,3)
    
    vg_norm = sqrt(ug**2 + vg**2)
    do k = 1, num_level
       sigma(:, :)  = p_full(:,:,k) / p_half(:,:,num_level+1)
       sigma_norm   = (sigma - sigma_b) / (1.0 - sigma_b)
       sigma_max    = max(sigma_norm, 0.0)
       dt_ug(:,:,k) = dt_ug(:,:,k) - Cdrag * sigma_max * vg_norm(:,:,k)*ug(:,:,k)
       dt_vg(:,:,k) = dt_vg(:,:,k) - Cdrag * sigma_max * vg_norm(:,:,k)*vg(:,:,k)
    end do
    
  end subroutine surface_drag

  !-----------------------------------------------------------------

  subroutine diabatic_forcing(Time, is, js, tg, p_half, p_full, lat, dt_tg)
     
    implicit none

    integer, intent(in)                   :: is, js
    type(time_type), intent(in)           :: Time
    ! temperature and pressure fields
    real, intent(in), dimension(:, :, :)  :: tg, p_full
    real, intent(in), dimension(:, :, :)  :: p_half 
    real, intent(in), dimension(:, :)     :: lat
    
    ! time tendency in temperature field
    real, intent(inout), dimension(:,:,:) :: dt_tg 
            
    ! local variables
    real, dimension(size(tg, 1), size(tg, 2), size(tg, 3)) ::              &
         heating_rate_xtropics,                                            &
         heating_rate_tropics,                                             &
         diab_heating_rate
    real, dimension(size(tg, 1), size(tg, 2)) ::                           &
         sin_lat,                                                          &
         sin_lat_2,                                                        & 
         tropic_dist,                                                      &
         tg_sfc 

    integer :: j, k, num_level
    logical :: used


    ! pseudo-parameters
    num_level = size(tg, 3)        

    if (size(lat,1) .ne. size(tg,1) &
         .or. size(lat,2) .ne. size(tg,2)) then
       call error_mesg('diabatic_forcing',                             &
            'mismatched argument list',fatal)
       ! terminate execution
    end if

    ! get ``distribution of tropics'': tropic_dist is equal to 1 at the 
    ! equator and rapidly decays to zero away from the equator
    sin_lat_2             = sin(lat)**2
    if(tropical_heating) then
       tropic_dist           = tropics_distribution(sin_lat_2)
    else
       tropic_dist = 0.0
    endif

    if (gray_atmosphere) then
       ! get heating rates in extratropics (from gray atmosphere model)
! fix these if they are to be used
!!$       tg_sfc                = surface_temperature_forced(sin_lat_2)
!!$       heating_rate_xtropics = heating_rate_gray(tg, tg_sfc, p_half)
    else
       heating_rate_xtropics = relax_to_gray_equilibrium(Time, tg, lat,      &
            p_full, p_half, is, js)
    endif

    ! get heating rate in tropics by relaxation towards moist adiabat
    ! (approximate in situ temperature as lower boundary condition)
    if(fix_tropic_surface_temperature) then
       tg_sfc = fix_tropic_temp
    else
       tg_sfc                = surface_temperature(tg, p_full, p_half)
    endif
    
    if(tropical_heating) then
       heating_rate_tropics  = relax_to_pseudoadiabat(tg, tg_sfc,     &
            p_full, p_half(:, :, num_level+1))
    else
       heating_rate_tropics  = 0.0
    endif

    do k = 1, num_level
       ! total heating rate is composed of the relaxation rate towards 
       ! the pseudoadiabat in the tropics and the heating rate that 
       ! applies to the entire atmosphere, including the tropics  
       heating_rate_tropics(:, :, k)=tropic_dist*heating_rate_tropics(:, :, k)
    enddo

       diab_heating_rate = heating_rate_tropics + heating_rate_xtropics
       dt_tg     = dt_tg + diab_heating_rate

! send data to diagnostic manager
       if(id_tdt > 0) used = send_data(id_tdt, diab_heating_rate, Time,      &
            is, js)

       if(id_tropical_tdt > 0) used = send_data(id_tropical_tdt,             &
            heating_rate_tropics, Time, is, js)

       if(id_extratropical_tdt > 0) used = send_data(id_extratropical_tdt,   &
            heating_rate_xtropics, Time, is, js)

  end subroutine diabatic_forcing

  !-----------------------------------------------------------------------
  
  function relax_to_gray_equilibrium(Time, temp, lat, p_full, p_half, is, js)&
       result (heating_rate)
  
    ! computes heating rate from relaxation towards radiative
    ! equilibrium state of a gray atmosphere with an isothermal 
    ! stratosphere 
    
    implicit none

    type(time_type), intent(in)           :: Time

    integer, intent(in) :: is, js

    real, intent(in), dimension(:, :) ::                                   &
         lat
    real, intent(in)                  ::                                   &
         temp(:, :, :),                                                    &
         p_full(:, :, :),                                                  &
         p_half(:, :, :)
    
    ! local variables
    real, dimension(size(temp, 1), size(temp, 2)) ::                       & 
         sin_lat_2,                                                        &
         cos_lat_2,                                                        &
         cos_lat_4,                                                        &
         cos_lat_8,                                                        & 
         sigma,                                                            &
         sigma_norm,                                                       &
         sigma_max,                                                        &
         kt,                                                               &
         t_sfc,                                                            &
         optical_thickness

    real, dimension(size(temp, 1), size(temp, 2), size(temp, 3)) ::        &
         p_full_ref,                                                       &
         temp_eq,                                                          &
         heating_rate

    integer :: num_level, k 
    logical used

    num_level = size(temp, 3)
    cos_lat_2 = cos(lat)*cos(lat)
    cos_lat_4 = cos_lat_2**2
    cos_lat_8 = cos_lat_2**4

    if (size(lat,1) .ne. size(p_full,1) &
         .or. size(lat,2) .ne. size(p_full,2)) then
       call error_mesg('relax_to_gray_equilibrium',                        &
            'mismatched argument list',fatal)
       ! terminate execution
    end if

    ! normalized full level pressure and functions thereof
    p_full_ref        = p_full / reference_sea_level_press

    ! surface temperature in background state    
    t_sfc             = surface_temperature_forced(lat) 

    ! latitude-dependent optical thickness
    optical_thickness = (t_sfc / t_strat)**4 - 1.

    do k = num_level, 1, -1
       ! compute current sigma-level to get height-dependence of cooling coeff
       sigma      = p_full(:, :, k) / p_half(:, :, num_level+1)
       sigma_norm = (sigma - sigma_b) / (1.0 - sigma_b)
       sigma_max  = max(sigma_norm, 0.0) ! sigma_max = 0 outside `PBL'

       ! Newtonian damping coefficient
       ! changed by CW 2/24 for new runs
!       kt         = ka + (ks - ka) * sigma_max * cos_lat_8(:,:) 
       kt         = ka + (ks - ka) * sigma_max * cos_lat_4(:,:) 
       
       ! temperature in background state is radiative equilibrium 
       ! of a gray atmosphere (cf. Goody and Yung, pp. 392 ff.)
       temp_eq(:,:,k)    =                                                 &
            t_strat * (1. + optical_thickness                              &
                     * p_full_ref(:, :, k)**scale_height_ratio)**(1./4.)

       ! relax towards background state
       heating_rate(:, :, k) = ( temp_eq(:,:,k)  - temp(:, :, k) ) * kt
    end do
    
    if (id_teq > 0) used = send_data ( id_teq, temp_eq,  Time, is, js)

  end function relax_to_gray_equilibrium

  !-----------------------------------------------------------------------

  function tropics_distribution(sin_lat_2)

    ! computes ``distribution of tropics'' to smooth out transition between 
    ! model tropics and extratropics

    real, dimension(:, :), intent(in) :: sin_lat_2
    real, dimension(size(sin_lat_2, 1), size(sin_lat_2, 2)) ::      &
         tropics_distribution, stretched_sin_lat_2
    
    stretched_sin_lat_2  = sin(width_tropics_rad)**(-2) * sin_lat_2
    tropics_distribution = 1. - tanh(stretched_sin_lat_2)
    
  end function tropics_distribution

  !-----------------------------------------------------------------------

  function heating_rate_gray(temp, temp_sfc, p_half)

    ! computes the heating rate in a gray atmosphere

    implicit none

    real, intent(in) ::       &
         temp(:, :, :),       & ! temperature field
         temp_sfc(:, :),      & ! surface temperature (lower boundary condition)
         p_half(:, :, :)        ! half-level pressure

    real, dimension(size(temp, 1), size(temp, 2), size(temp, 3)) ::        &
         heating_rate_gray

    real, dimension(size(temp, 1), size(temp, 2), size(temp, 3)) ::        &
         source_function        ! source of thermal radiation (Planck function)
    real, dimension(size(temp, 1), size(temp, 2)) ::                       &
         source_function_sfc, & ! source function at surface (lower bdry cndtn)
         optical_thickness      ! optical thickness of atmosphere
    real, dimension(size(p_half, 1), size(p_half, 2), size(p_half, 3)) ::  &
         up_flux,             & ! upward flux of thermal radiation
         down_flux,           & ! downward flux of thermal radiation
         net_flux,            & ! total flux (difference of up and down fluxes)
         tau,                 & ! optical depth
         exp_tau,             & ! exp(optical depth)
         exp_tau_inv            ! exp(-optical depth)

    integer :: num_level, k

    num_level             = size(temp, 3)

    ! get approximate total optical thickness of atmosphere
    optical_thickness     = gray_optical_thickness(temp_sfc)
    
    ! get approximate optical depth and derived functionals
    tau(:, :, 1)          = 0.0   ! optical depth at upper boundary
    do k=2, num_level + 1
       tau(:, :, k)       = optical_thickness                              &
            * (p_half(:, :, k) / reference_sea_level_press)**scale_height_ratio 
    enddo
    exp_tau               = exp(tau)
    exp_tau_inv           = 1. / exp_tau

    source_function       = stefan * (temp**4)
    source_function_sfc   = stefan * (temp_sfc**4)

    down_flux(:, :, 1)    = 0.0      ! upper boundary condition for downward flux
    do k = 2, num_level+1            ! integrate flux downward
       down_flux(:, :, k) = down_flux(:, :, k-1)                           &
            + source_function(:,:, k-1) * (exp_tau(:,:, k) - exp_tau(:,:, k-1))
    end do
    down_flux             = down_flux * exp_tau_inv

    up_flux(:, :, num_level+1) = 0.0 ! integrate upward flux from lower boundary
    do k = num_level, 1, -1          ! upwards, adding lower boundary condition later
       up_flux(:, :, k)   = up_flux(:, :, k+1)                             &
            + source_function(:,:, k) * (exp_tau_inv(:,:, k) - exp_tau_inv(:,:, k+1))
    end do
    up_flux               = up_flux * exp_tau

    do k=1, num_level+1              ! add upward flux from lower boundary
       up_flux(:, :, k)   = up_flux(:, :, k)                               &
            + source_function_sfc * exp_tau(:, :, k) * exp_tau_inv(:, :, num_level+1)
    end do

    ! get total flux as difference between upward flux and downward flux
    net_flux              = up_flux - down_flux

    ! compute heating rate from vertical divergence of net flux
    do k =1, num_level
       heating_rate_gray(:, :, k) = grav / cp_air &
            * ( net_flux(:, :, k+1) - net_flux(:, :, k) )                    &
            / ( p_half(:, :, k+1) - p_half(:, :, k) )
    end do

  end function heating_rate_gray

  !-------------------------------------------------------------------------
  
  function gray_optical_thickness(temp_sfc) result(optical_thickness)

    ! Parametric expression for total optical thickness: The total 
    ! optical thickness is assumed to be proportional the the 
    ! saturation vapor pressure over the surface. As the saturation 
    ! vapor pressure is temperature-dependent, the optical thickness
    ! depends on the surface temperature and hence on latitude. However, 
    ! since in the polar regions the total optical thickness of the atmosphere 
    ! is larger than the optical thickness due to water vapor alone, the
    ! optical thickness is constrained not to become smaller than a minimum
    ! value min_optical_thickness. (The optical thickness due to water vapor 
    ! alone is nearly zero in polar regions, which, if considered
    ! alone, would result in excessively strong radiative cooling near
    ! the ground.) 

    real, intent(in), dimension(:, :) :: temp_sfc

    real, dimension(size(temp_sfc, 1), size(temp_sfc, 2)) ::               &
         sat_vapor_press,     & ! saturation vapor pressure over surface
         optical_thickness

    real, parameter, dimension(1, 1) ::        &
         sin_lat_2_eq     = 0.
    real, dimension(1, 1) ::                   &
         temp_eq,             & ! temperature at equator
         sat_vapor_press_eq     ! saturation vapor pressure at equator
    
    integer :: j

    ! surface temperature at equator
    temp_eq               = surface_temperature_forced(sin_lat_2_eq)
    ! corresponding saturation vapor pressure
    sat_vapor_press_eq    = saturation_vapor_pressure(temp_eq)

    ! saturation vapor pressure everywhere at surface
    sat_vapor_press       = saturation_vapor_pressure(temp_sfc)
  
    ! optical thickness
    optical_thickness     = max_optical_thickness *                        &
         sat_vapor_press / sat_vapor_press_eq(1,1)
    optical_thickness     = max(optical_thickness, min_optical_thickness)

  end function gray_optical_thickness

  !-------------------------------------------------------------------------

  function relax_to_pseudoadiabat(temp, temp_sfc, p_full, p_sfc)           &
       result (heating_rate)

    ! computes the heating rate as Newtonian relaxation rate towards a
    ! pseudoadiabat

    implicit none
    
    real, parameter ::                                                     &
         eps_heat = rdgas/rvgas

    real, parameter ::                                                     &
         specific_heat_watervap_pressure = 1952.0

    real, intent(in) ::                                                    &
         temp(:, :, :),    & ! in situ temperature field
         temp_sfc(:, :),   & ! surface temperature (lower boundary condition)
         p_full(:, :, :),  & ! full-level pressure
         p_sfc(:, :)         ! surface pressure
    
    real, dimension(size(temp, 1), size(temp, 2), size(temp, 3)) ::        &
         heating_rate

    real, dimension(size(temp, 1), size(temp, 2)) ::                       &
         pseudoadiabat_temp,                                               &
         pseudoadiabat_temp_below,                                         &
         pressure_below,                                                   &
         sat_vapor_press,                                                  &
         sat_mix_rat,                                                      &
         pseudoadiabat_lapse_rate,                                         &
         c_p_moist,                                                        &
         gamma

    integer :: num_level, k
    num_level                = size(temp, 3)

    pseudoadiabat_temp_below = temp_sfc
    pressure_below           = p_sfc
    do k = num_level, 1, -1 
       ! integrate pseudoadiabatic lapse rate from the surface upwards:
        
       ! compute the saturation vapor pressure at level below
       sat_vapor_press = saturation_vapor_pressure(pseudoadiabat_temp_below)

       ! get saturation mixing ratio
       sat_mix_rat     = eps_heat * sat_vapor_press                        &
            / ( pressure_below - sat_vapor_press )

       ! dimensionless quantity appearing in lapse rate
       gamma           = 1. + sat_mix_rat / eps_heat

       ! get moisture corrected specific heat of air
       c_p_moist       = cp_air                    &
            + sat_mix_rat * specific_heat_watervap_pressure

       ! get *pressure* lapse rate on half level below
       ! (modified from Emmanuel p. 131)
       pseudoadiabat_lapse_rate = gamma / pressure_below                   &
            * ( rdgas * pseudoadiabat_temp_below            &
                + hlv * sat_mix_rat )                      &
            / ( c_p_moist                                                  &
                + hlv**2 * sat_mix_rat * gamma             &
                  / rvgas / pseudoadiabat_temp_below**2 )

       ! rescale pseudoadiabatic lapse rate by scale factor
       pseudoadiabat_lapse_rate = lapse_scale_tropic * pseudoadiabat_lapse_rate

       pseudoadiabat_temp       = pseudoadiabat_temp_below                 &
            + pseudoadiabat_lapse_rate * (p_full(:, :, k) - pressure_below)
          
       ! heating rate as relaxation towards moist adiabat up to level of 
       ! neutral buoyancy; heating rate is zero above level of neutral
       ! buoyancy 
       where (pseudoadiabat_temp .gt. temp(:, :, k))
          heating_rate(:, :, k)    =                                       &
               kt_tropic * (pseudoadiabat_temp - temp(:, :, k))
       elsewhere
          heating_rate(:, :, k)    = 0.
       endwhere

       pseudoadiabat_temp_below = pseudoadiabat_temp
       pressure_below           = p_full(:, :, k)
    enddo

  end function relax_to_pseudoadiabat

  !--------------------------------------------------------------------------
  
  function saturation_vapor_pressure(temp_xy)

    ! computes the saturation vapor pressure at a given temperature 
    ! temp_xy (which must be measured in units of K)
    ! uses a formula due to Bolton (1980)
    
    implicit none
    
    real, dimension(:, :), intent(in) :: temp_xy
    
    real, dimension(size(temp_xy, 1), size(temp_xy, 2)) ::                 &
         temp_ratio,                                                       &
         saturation_vapor_pressure

    temp_ratio                = 17.67 * (temp_xy - tfreeze)           &
         / (temp_xy - 29.66)
    saturation_vapor_pressure = 611.2 * exp(temp_ratio)
    
  end function saturation_vapor_pressure

  !--------------------------------------------------------------------------

  function surface_temperature(temp, p_full, p_half)

    ! estimate the surface temperature from the temperature on lowest
    ! full level by assuming that the potential temperature on the surface
    ! is equal to the potential temperature on the lowest full level

    implicit none

    real, intent(in) ::                                                    &
         temp(:, :, :),    & ! temperature field
         p_full(:, :, :),  & ! full-level pressure
         p_half(:, :, :)     ! half-level pressure

    real, dimension(size(temp, 1), size(temp, 2)) ::                       &
         surface_temperature

    integer :: num_level
    num_level = size(temp, 3)

    surface_temperature = temp(:, :, num_level)                            &
         * ( p_half(:, :, num_level+1) / p_full(:, :, num_level) )**kappa
    
  end function surface_temperature
  
  !--------------------------------------------------------------------------

  function surface_temperature_forced(lat)

    ! returns a prescribed surface temperature as forcing on lower boundary

    implicit none

    real, intent(in) ::                                                    &
         lat(:, :)
    real, dimension(size(lat, 1), size(lat, 2)) ::                         &
         surface_temperature_forced,                                       &
         legendre_polynomial_2,                                            &
         sin_lat_2

    real phi_not_2

    sin_lat_2 = sin(lat)*sin(lat)


    legendre_polynomial_2 = 1. / 3. - sin_lat_2  ! Second Legendre polynomial
    surface_temperature_forced = tsfc_mean + delh * legendre_polynomial_2 

    surface_temperature_forced = surface_temperature_forced + off_eq*sin(lat)
!!$
!!$    surface_temperature_forced = tsfc_mean -                               &
!!$         delh * (0.5 * (3.*alpha-1.) * sin_lat_2 - 1.5 * (alpha-1.) * sin_lat_2**2)
  
  end function surface_temperature_forced
  
  !--------------------------------------------------------------------------

  function axisymmetric(field_xy)

    ! averages xy-field longitudinally and returns it as axisymmetric field

    implicit none

    real, intent(in) ::                                                    &
         field_xy(:, :)
    real, dimension(size(field_xy, 1), size(field_xy, 2)) ::               &
         axisymmetric
    real, dimension(size(field_xy, 2)) ::                                  &
         zonal_mean
    integer :: j

    zonal_mean = sum(field_xy, 1) / float( size(field_xy, 1) )
    do j=1, size(field_xy, 2)
       axisymmetric(:, j) = zonal_mean(j)
    enddo

  end function axisymmetric

  
!--------------------------------------------------------------------------

!#######################################################################
 subroutine hs_forcing_end 

!-----------------------------------------------------------------------
!
!       routine for terminating held-suarez benchmark module
!             (this routine currently does nothing)
!
!-----------------------------------------------------------------------


 end subroutine hs_forcing_end

!#######################################################################

 subroutine hs_newtonian_damping ( lat, ps, p_full, t, tdt, teq, mask )

!-----------------------------------------------------------------------
!
!   routine to compute thermal forcing for held & suarez (1994)
!   benchmark calculation.
!
!-----------------------------------------------------------------------

real, intent(in),  dimension(:,:)   :: lat, ps
real, intent(in),  dimension(:,:,:) :: p_full, t
real, intent(out), dimension(:,:,:) :: tdt, teq
real, intent(in),  dimension(:,:,:), optional :: mask

!-----------------------------------------------------------------------

          real, dimension(size(t,1),size(t,2)) :: &
     sin_lat, sin_lat_2, cos_lat_2, t_star, cos_lat_4, &
     tstr, sigma, the, tfactr, rps, p_norm

       real, dimension(size(t,1),size(t,2),size(t,3)) :: tdamp

       integer :: k
       real    :: tcoeff, pref

!-----------------------------------------------------------------------
!------------latitudinal constants--------------------------------------

      sin_lat  (:,:) = sin(lat(:,:))
      sin_lat_2(:,:) = sin_lat(:,:)*sin_lat(:,:)
      cos_lat_2(:,:) = 1.0-sin_lat_2(:,:)
      cos_lat_4(:,:) = cos_lat_2(:,:)*cos_lat_2(:,:)

      t_star(:,:) = tsfc_mean - delh*sin_lat_2(:,:)
      tstr  (:,:) = t_strat 

!-----------------------------------------------------------------------

      tcoeff = (ks-ka)/(1.0-sigma_b)
      pref = reference_sea_level_press
      rps  = 1./ps

      do k = 1, size(t,3)

!  ----- compute equilibrium temperature (teq) -----
         p_norm(:,:) = p_full(:,:,k)/pref
         the   (:,:) = t_star(:,:) - delv*cos_lat_2(:,:)*log(p_norm(:,:))
         teq(:,:,k) = the(:,:)*(p_norm(:,:))**KAPPA
         teq(:,:,k) = max( teq(:,:,k), tstr(:,:) )

!  ----- compute damping -----
         sigma(:,:) = p_full(:,:,k)*rps(:,:)
         where (sigma(:,:) <= 1.0 .and. sigma(:,:) > sigma_b)
           tfactr(:,:) = tcoeff*(sigma(:,:)-sigma_b)
           tdamp(:,:,k) = ka + cos_lat_4(:,:)*tfactr(:,:)
         elsewhere
           tdamp(:,:,k) = ka
         endwhere

      enddo

!*** note: if the following loop uses vector notation for all indices
!          then the code will not run ??????

      do k=1,size(t,3)
         tdt(:,:,k) = -tdamp(:,:,k)*(t(:,:,k)-teq(:,:,k))
      enddo

      if (present(mask)) then
         tdt = tdt * mask
         teq = teq * mask
      endif

!-----------------------------------------------------------------------

 end subroutine hs_newtonian_damping

end module grid_physics


